////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Host Communications Channel Module Internal
///
///     This file defines the internal HCC module function prototypes and
///     definitions.
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include "tera_pcoip_hcc_apdu.h"
#include "tera_mem.h"
#include "tera_pkt.h"
#include "tera_scp.h"

// ---- Constants and Types -------------------------------------------------

// --------------------------------------------------------------------------
//  sTERA_HCC_MSG_IOCTL_RSP
///     ioctl response message
//
//  Description:
///     This union includes all of the possible responses to ioctl commands.
///     When an ioctl command is received that requires a response from the
///     firmware the thread executing the ioctl command waits for the response
///     to be received by the dpc_entry from the firmware. The dpc_entry
///     copies the response into the control block and wakes the waiting
///     thread that is servicing the icotl commands. The ioctl thread then
///     copies the data from the control block into the ioctl buffer.
///
typedef struct sTERA_HCC_MSG_IOCTL_RSP_STRUCT
{
    union
    {
        sTERA_PCOIP_HCC_APDU_HMI_VAR               get_rep;            ///< Get reply
        sTERA_PCOIP_HCC_APDU_HMI_VAR               set_req;            ///< Set request
        sTERA_PCOIP_HCC_APDU_KMP_PRI_CTXT          kmp_pri;            ///< KMP PRI Context
        sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO         img_disp_topo;      ///< IMG Display Topology
        sTERA_PCOIP_HCC_APDU_IMG_CACHE_DISP_TOPO   img_cache_topo_req; ///< IMG Cached Topo request
        sTERA_PCOIP_HCC_APDU_IMG_USER_CFG_REQ_DATA usr_cfg;            ///< IMG User Config settings
    } data;
} sTERA_HCC_MSG_IOCTL_RSP;

// --------------------------------------------------------------------------
//  sHCC_HMI_CBLK
///     HMI user control block
//
//  Description:
///     This structure defines the HMI user control block. The control block is
///     per device instance, and not per driver instance. The control block is
///     attached to the context area of each device instance.
///
typedef struct sHCC_HMI_CBLK_STRUCT
{
    struct semaphore                       data_lock;      ///< Data spinlock
    sTERA_PCOIP_HCC_APDU_HMI_MAP           avail_cmd;      ///< Available commands
    sTERA_PCOIP_HCC_APDU_HMI_MAP           avail_var_get;  ///< Available variables for getting
    sTERA_PCOIP_HCC_APDU_HMI_MAP           avail_var_set;  ///< Available variables for setting
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;      ///< Session link
} sHCC_HMI_CBLK;

// --------------------------------------------------------------------------
//  sHCC_CBLK
///     HCC module control block
//
//  Description:
///     This structure defines the Host Communications Channel (HCC) module
///     control block. The control block is per device instance, and not per
///     driver instance. The control block is attached to the context area of
///     each device instance.
///
typedef struct sHCC_CBLK_STRUCT
{
    struct pci_dev *dev;                 ///< Device object
    INT             initialized;         ///< Initialized flag
    INT             started;             ///< Started flag
    TERA_SCP_HANDLE scp_handle;          ///< SCP handle

    wait_queue_head_t pending_ioctl_q;   ///< Pending ioctl queue
    struct semaphore pending_ioctl_lock; ///< Pending ioctl semaphore
    INT32   pending_ioctl_flag;          ///< flag used to signal the FW req completed
    sTERA_HCC_MSG_IOCTL_RSP ioctl_rsp;   ///< copy of data from FW that belongs in the ioctl response

    struct semaphore data_lock;          ///< Local data semaphore

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
    struct delayed_work work;            ///< dpc_entry work queue task
#else
    struct work_struct work;             ///< dpc_entry work queue task
#endif

    sTERA_PKT_Q     egress_pkt_q;        ///< Egress packet queue
    sHCC_HMI_CBLK   hmi;                 ///< HMI user control block
    sTERA_PCOIP_HCC_APDU_IMG_DISP_TOPO    async_disp_topo;
                                         ///< Last received asynchronous display topology
    UINT32          id_disp_topo;        ///< ID of last async topo APDU
	eTERA_PCOIP_HMI_HOST_SESSION_STATE    host_sess_state;
                                         ///< host session state
    sTERA_PCOIP_HCC_IMG_MODE              user_cfg_setting;
                                         ///< imaging user config settings
} sHCC_CBLK;

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------

//
// HMI user function prototypes
//

extern INT32 hcc_hmi_init(
    sHCC_CBLK *cblk
    );

extern INT32 hcc_hmi_start(
    sHCC_CBLK *cblk
    );

extern void hcc_hmi_stop(
    sHCC_CBLK *cblk
    );

extern INT32 hcc_hmi_ioctl_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

extern void hcc_hmi_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

extern void hcc_hmi_scp_chan_open_proc(
    sHCC_CBLK *cblk
    );

//
// KMP user function prototypes
//

extern INT32 hcc_kmp_init(
    sHCC_CBLK *cblk
    );

extern INT32 hcc_kmp_start(
    sHCC_CBLK *cblk
    );

extern void hcc_kmp_stop(
    sHCC_CBLK *cblk
    );

extern INT32 hcc_kmp_ioctl_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

extern void hcc_kmp_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

//
// IMG user function prototypes
//

extern INT32 hcc_img_init(
    sHCC_CBLK *cblk
    );

extern INT32 hcc_img_ioctl_proc(
    sHCC_CBLK    *cblk,
    unsigned long arg,
    unsigned int  ioctl
    );

extern void hcc_img_apdu_proc(
    sHCC_CBLK *cblk,
    void      *apdu,
    UINT32     apdu_len
    );

