////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     Message Module Implementation
///
///     This file implements the message module, which implements the message
///     queues, queue entries, and operations on the queues.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include "tera_log.h"
#include "tera_pcoip_driver.h"
#include "tera_errors.h"
#include "tera_msg.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// ---- Local Variables -----------------------------------------------------

// Message pool name string table
static const char *msg_pool_name_str_tbl[MAX_NUM_GFUN] = {
    "pcoipHostMsgPool0",
    "pcoipHostMsgPool1"
};

// ---- Local Function Prototypes -------------------------------------------

static sTERA_MSG *msg_alloc(
    sMSG_CBLK *cblk
    );

// --------------------------------------------------------------------------
//  msg_alloc
//      Allocate one message
//
//  Description:
//      This function allocate one message with the specific buffer length. The
//      memory for the message header and buffer is allocated from a lookaside
//      list to keep it efficient. If there is not enough memory then this
//      function returns NULL.
//
static sTERA_MSG *msg_alloc(
    sMSG_CBLK *cblk       // Control block
    )
{
    sTERA_MSG *ret = NULL;

    ASSERT(cblk != NULL);

    // Create a new buffer from lookaside list for the message structure
    ret = kmem_cache_alloc(cblk->msg_mem_pool, GFP_KERNEL);
    if (!ret)
    {
        mTERA_LOG_ERR("kmem_cache_alloc failed");
    }
    else
    {
        // Zero the message memory
        memset(ret, 0, sizeof(sTERA_MSG) + cblk->buf_len);

        ret->handle = ret;
        ret->buf_len = cblk->buf_len;

        // Increment the message allocate counter
        cblk->msg_alloc_cnt++;
    }

    return ret;
} // msg_alloc()

// --------------------------------------------------------------------------
//  tera_msg_init
///     Initialize the message module
//
//  Description:
///     This function initializes the message module. The user must specify
///     the size of all the messages that will be allocated from this module.
///     It is up to the user to make sure all message allocation requests to
///     this module will NOT exceed the size specified.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_NO_MEMORY
///
INT32 tera_msg_init(
    sMSG_CBLK *cblk,     ///< Message control block
    UINT32     buf_len,  ///< Length of each message buffer to be allocated
    UINT32     dev_id    ///< Device ID
    )
{
    INT32                 ret = TERA_ERR_FAILURE;

    ASSERT(buf_len > 0);
    ASSERT(dev_id < MAX_NUM_GFUN);

    // Create a lookaside list for buffers for internal messages
    cblk->msg_mem_pool = kmem_cache_create(msg_pool_name_str_tbl[dev_id],
                                           sizeof(sTERA_MSG) + buf_len,
                                           0, SLAB_HWCACHE_ALIGN,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23))
                                           NULL);
#else
                                           NULL, NULL);
#endif
    if (!cblk->msg_mem_pool)
    {
        ret = TERA_ERR_NO_MEMORY;
        mTERA_LOG_CRIT("kmem_cache_create failed");
    }
    else
    {
        // Record the message size
        cblk->buf_len = buf_len;

        // Initialize the control block fields
        cblk->initialized = 1;
        cblk->msg_alloc_cnt = 0;
        cblk->msg_free_cnt = 0;
        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_msg_init()

// --------------------------------------------------------------------------
//  tera_msg_q_init
///     Initialize a message queue
//
//  Description:
///     This function initializes a message queue by creating the spinlock
///     that is used for accessing the queue, and also initializing the head
///     pointer. A message queue MUST be initialized before performing enqueuing
///     or dequeuing messages to/from it.
///
/// \return
///     TERA_SUCCESS,
///     TERA_ERR_FAILURE
///
INT32 tera_msg_q_init(
    sTERA_MSG_Q *msg_q    ///< message queue
    )
{
    INT32   ret = TERA_ERR_INVALID_ARG;

    if (msg_q != NULL)
    {
        // Initialize the message queue variables
        init_MUTEX(&msg_q->lock);
        msg_q->head = NULL;
        msg_q->initialized = 1;
        msg_q->enqueue_cnt = 0;
        msg_q->dequeue_cnt = 0;

        ret = TERA_SUCCESS;
    }

    return ret;
} // tera_msg_q_init()

// --------------------------------------------------------------------------
//  tera_msg_cleanup
///     Clean up the message module
//
//  Description:
///     This function cleans up the message module. This function marks the
///     module as uninitialized. All operations on messages and message queues,
///     such as freeing or purging, should be done before calling this
///     function.
///
/// \return
///     None
///
void tera_msg_cleanup(
    sMSG_CBLK *cblk       ///< Message control block
    )
{
    mTERA_LOG_DBG("start");

    if ((cblk != NULL) && cblk->initialized)
    {
        mTERA_LOG_INFO("msg_alloc_cnt=%u, msg_free_cnt=%u",
                       cblk->msg_alloc_cnt, cblk->msg_free_cnt);

        // Mark the module as uninitialized
        cblk->initialized = 0;

        // destroy the lookaside cache
        kmem_cache_destroy(cblk->msg_mem_pool);
    }
    else
    {
        mTERA_LOG_ERR("Unexpected state or invalid parameter");
    }

    mTERA_LOG_DBG("end");
} // tera_msg_cleanup()

// --------------------------------------------------------------------------
//  tera_msg_alloc
///     Allocate one or more messages
//
//  Description:
///     This function allocates one or more messages with the specific buffer
///     length. If there is not enough memory for the requested number of
///     messages, then no messages will be returned. If successful, the
///     requested number of messages are returned, chained one after another
///     via the next pointer. The message buffers are zeroed out.
///
///     If the buffer length is greater than the size specified in the init
///     function, then this function returns NULL.
///
/// \return
///     Pointer to a message,
///     NULL
///
sTERA_MSG *tera_msg_alloc(
    struct pci_dev *dev,    ///< Device object
    UINT32    buf_len,      ///< Message buffer length
    UINT32    num_msg       ///< Number of messages to allocate
    )
{
    sTERA_DEV_CBLK *dev_cblk;
    sTERA_MSG      *ret = NULL;
    sTERA_MSG     **msg;
    UINT32          i;

    if (dev == NULL)
    {
        mTERA_LOG_ERR("invalid state");
        return ret;
    }
    dev_cblk = pci_get_drvdata(dev);

    if((dev_cblk == NULL) || (!dev_cblk->initialized))
    {
        mTERA_LOG_ERR("device control block NULL or Not Initialized");
    }
    else if (buf_len > dev_cblk->msg_cblk.buf_len)
    {
        mTERA_LOG_ERR("invalid message length (requested=%u, max=%u)",
                      buf_len, dev_cblk->msg_cblk.buf_len);
    }
    else
    {
        msg = &ret;

        for (i = 0; i < num_msg; i++)
        {
            *msg = msg_alloc(&dev_cblk->msg_cblk);

            if (*msg == NULL)
            {
                mTERA_LOG_ERR("failed to allocate message %u of %u", i + 1,
                              num_msg);

                // Free all the buffers
                tera_msg_free(dev, ret);

                ret = NULL;
                break;
            }

            msg = &(*msg)->next;
        }
    } // end if

    return ret;
} // tera_msg_alloc()

// --------------------------------------------------------------------------
//  tera_msg_free
///     Free the memory used by the message(s)
//
//  Description:
///     This function frees the memory that was used by the message(s). The
///     caller can pass in multiple messages by chaining the messages via the
///     next pointer.
///
/// \return
///     None
///
void tera_msg_free(
    struct pci_dev *dev,    ///< Device object
    sTERA_MSG *msg          ///< Message
    )
{
    sTERA_DEV_CBLK *dev_cblk;
    sMSG_CBLK *cblk;
    sTERA_MSG *curr;
    sTERA_MSG *next;

    ASSERT(dev != NULL);

    dev_cblk = pci_get_drvdata(dev);
    cblk = &dev_cblk->msg_cblk;

    if (cblk->initialized && (msg != NULL))
    {
        curr = msg;

        while (curr != NULL)
        {
            next = curr->next;

            // Delete the memory object of which the buffer belongs to
            kmem_cache_free(cblk->msg_mem_pool, curr->handle);

            // Increment the message free counter
            cblk->msg_free_cnt++;

            // Move on to the next message
            curr = next;
        }
    }
} // tera_msg_free()

// --------------------------------------------------------------------------
//  tera_msg_q_purge
///     Purge all messages from the message queue
//
//  Description:
///     This function purges all messages from the message queue. This function
///     asserts if the message queue is NULL or not initialized.
///
void tera_msg_q_purge(
    struct pci_dev *dev,  ///< Device object
    sTERA_MSG_Q *msg_q    ///< message queue
    )
{
    sMSG_CBLK      *cblk;
    sTERA_DEV_CBLK *dev_cblk;
    sTERA_MSG      *msg;
    UINT32          purge_cnt = 0;

    ASSERT(msg_q != NULL);
    ASSERT(msg_q->initialized);
    ASSERT(dev != NULL);

    dev_cblk = pci_get_drvdata(dev);
    cblk = &dev_cblk->msg_cblk;

    while ((msg = tera_msg_q_dequeue(msg_q)) != NULL)
    {
        // Free the message
        tera_msg_free(dev, msg);

        purge_cnt++;
    }

    mTERA_LOG_DBG("enqueue_cnt=%u, dequeue_cnt=%u",
                  msg_q->enqueue_cnt, msg_q->dequeue_cnt);

    mTERA_LOG_DBG("purged %u messages", purge_cnt);
} // tera_msg_q_purge()

// --------------------------------------------------------------------------
//  tera_msg_q_enqueue
///     Enqueue one or more messages to the tail of the message queue
//
//  Description:
///     This function enqueues one or more messages to the tail of the specified
///     message queue. The user can enqueue more than one message by chaining
///     the messages using the next pointer. The last message in the chain MUST
///     have a NULL next pointer. This function is synchronized to other access
///     attempts to the message queue. This function asserts if the message
///     queue is NULL or not initialized.
///
/// \return
///     None
///
void tera_msg_q_enqueue(
    sTERA_MSG_Q *msg_q,    ///< Message queue
    sTERA_MSG   *msg       ///< Message
    )
{
    sTERA_MSG *node;

    ASSERT(msg_q != NULL);
    ASSERT(msg_q->initialized);

    if (msg != NULL)
    {
#ifdef DBG
        // Count how many messages we are going to enqueue
        node = msg;

        do
        {
            msg_q->enqueue_cnt++;

            node = node->next;
        } while (node != NULL);
#endif

        down(&msg_q->lock);

        if (msg_q->head == NULL)
        {
//            mTERA_LOG_DBG("enqueuing to head of message queue");

            msg_q->head = msg;
        }
        else
        {
            node = msg_q->head;

//            mTERA_LOG_DBG("enqueuing to tail of message queue");

            while (node->next != NULL)
            {
                node = node->next;
            }

            node->next = msg;
        }

//        mTERA_LOG_DBG("enqueue_cnt=%u", msg_q->enqueue_cnt);

        up(&msg_q->lock);
    }
} // tera_msg_q_enqueue()

// --------------------------------------------------------------------------
//  tera_msg_q_head_enqueue
///     Enqueue one or more messages to the head of the message queue
//
//  Description:
///     This function enqueues one or more messages to the head of the specified
///     message queue. The user can enqueue more than one message by chaining
///     the messages using the next pointer. The last message in the chain MUST
///     have a NULL next pointer. This function is synchronized to other access
///     attempts to the message queue. This function asserts if the message
///     queue is NULL or not initialized.
///
/// \return
///     None
///
void tera_msg_q_head_enqueue(
    sTERA_MSG_Q *msg_q,    ///< Message queue
    sTERA_MSG   *msg       ///< Message
    )
{
    sTERA_MSG *last = msg;

    ASSERT(msg_q != NULL);
    ASSERT(msg_q->initialized);

    if (msg != NULL)
    {
        for (;;)
        {
#ifdef DBG
            msg_q->enqueue_cnt++;
#endif

            if (last->next == NULL)
            {
                // We've reached the last message
                break;
            }

            last = last->next;
        }

        down(&msg_q->lock);

        if (msg_q->head == NULL)
        {
            msg_q->head = msg;
        }
        else
        {
            //
            // Set the last message in the chain to point to the current message
            // at the head, and then point the head to the start of the message
            // chain
            //
            last->next = msg_q->head;
            msg_q->head = msg;
        }

//        mTERA_LOG_DBG("enqueue_cnt=%u", msg_q->enqueue_cnt);

        up(&msg_q->lock);
    }
} // tera_msg_q_head_enqueue()

// --------------------------------------------------------------------------
//  tera_msg_q_dequeue
///     Dequeue a message from the head of the specified message queue
//
//  Description:
///     This function dequeues a message from the head of the specified message
///     queue. If the message queue is empty, this function returns NULL.
///
///     This function may be called after this module has been initialized. It
///     does not care if the module has been started or stopped.
///
/// \return
///     Pointer to a message,
///     NULL
///
sTERA_MSG *tera_msg_q_dequeue(
    sTERA_MSG_Q *msg_q    ///< Message queue
    )
{
    sTERA_MSG *ret = NULL;

    ASSERT(msg_q != NULL);
    ASSERT(msg_q->initialized);

    if(!down_interruptible(&msg_q->lock))
    {
        if (msg_q->head != NULL)
        {
            //
            // Take the message off of the head of the message queue and detach it
            // from the queue by assigning its next pointer to NULL
            //
            ret = msg_q->head;
            msg_q->head = (msg_q->head)->next;
            ret->next = NULL;
#ifdef DBG
            msg_q->dequeue_cnt++;
#endif
        }
#ifdef DBG
//        mTERA_LOG_DBG("dequeue_cnt=%u", msg_q->dequeue_cnt);
#endif
        up(&msg_q->lock);
    }
    return ret;
} // tera_msg_q_dequeue()

