////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Device Driver Interface
///
///     This file defines the PCoIP device driver interface and public ioctl
///     codes. Third party applications can use these defines to talk to the
///     PCoIP device and obtain useful information such as the host card MAC
///     address.
///
///     Requirements
///     ------------
///     The use of the PCoIP device interface requires the following:
///     - TERA1 PCoIP Host device must be present on the PCI bus
///     - Host Driver Function feature must be enabled on the PCoIP Host device
///     - PCoIP Host Software supporting third party application interaction
///       must be installed and running
///
///     Usage Guideline
///     ---------------
///     1. Write an application that opens the file descriptor associated with
///        the PCoIP Host driver. Typically the decriptor is named pcoip_host0
///        and can be found in the /dev directory. If the file descriptor is
///        not present either the driver is not installed or no devices are
///        present.
///     2. Use the ioctl API to retrieve information from the device. Also,
///        use the defined structures, if any, for the ioctl input/output
///        buffers, e.g.,
///
///             #define PCOIP_FILE "/dev/pcoip_host0"
///
///             sTERA_PCOIP_IOCTL_HOST_MAC   mac_msg;
///             int fd;
///             int ret;
///
///             fd = open(PCOIP_FILE, O_RDWR);
///             if (fd < 0)
///             {
///                 fprintf(stderr, "Failed to open PCoIP Host device");
///                 return 1;
///             }
///             ret = ioctl(fd, TERA_PCOIP_IOCTL_HOST_MAC_GET, &mac_msg);
///             printf("Host card MAC address: '%s'\n", mac_msg.mac);
///
///             close(fd);
///
////////////////////////////////////////////////////////////////////////////

#pragma once

// ---- Include Files -------------------------------------------------------

#include <linux/ioctl.h>

// ---- Constants and Types -------------------------------------------------

//
// In Linux the IOCTL command is 32 bits wide and the driver is allowed to
// define 16 bits of the command. Windows allows the driver to define 11 bits.
// To keep things simple the Linux version of the driver should not make use of
// more than 11 bits. The table below shows how the 16 bits are partitioned.
//
// All ioctls are in the range 0x7800 - 0x79FF with the following format:
//       Bit(s) | Description
//      --------|--------------------------------------------------
//      11-15   | Constant (0x78)
//       8-10   | Device interface ID (HCC or HDC)
//       5-7    | Device interface-specific user ID
//       0-4    | User-specific command/type
//

// List of device interface IDs
#define TERA_PCOIP_DEV_HCC_DEVIF  (0x0)
#define TERA_PCOIP_DEV_HDC_DEVIF  (0x1)

// Macro to define the upper 5 bits of the type field in the IOCTL cmd
#define TERA_PCOIP_TYPE_OFFSET    (0x78)

// Macro to define IOCTL codes. The supplied parameter 'n' is 8 bits wide and
// it gets mapped to the lower 8 bits of the IOCTL command.
#define mTERA_PCOIP_HCC_IOCTL(n)  _IO((TERA_PCOIP_TYPE_OFFSET + TERA_PCOIP_DEV_HCC_DEVIF), n)
#define mTERA_PCOIP_HDC_IOCTL(n)  _IO((TERA_PCOIP_TYPE_OFFSET + TERA_PCOIP_DEV_HDC_DEVIF), n)

// Macros to get the device interface and user ID from an ioctl code
#define mTERA_PCOIP_DEV_IOCTL_DEVIF_ID_GET(ioctl)  (((ioctl) & 0x0700) >> 8)
#define mTERA_PCOIP_DEV_IOCTL_USER_ID_GET(ioctl)   (((ioctl) & 0x00E0) >> 5)
#define mTERA_PCOIP_DEV_IOCTL_USER_CMD_GET(ioctl)   ((ioctl) & 0x001F)

//
// Ioctl data structures
//

// ---- Variable Externs ----------------------------------------------------

// ---- Function Prototypes -------------------------------------------------
