////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Daemon
///
///     This file implements the PCoIP Host SW daemon (pcoiphostswd). This
///     daemon runs when the workstation boots. The daemon is responsible for
///     enabling the local cursor feature in relative mode for SC sessions
///     when a user is not logged in locally.
///
////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <syslog.h>
#include <errno.h>
#include <pwd.h>
#include <signal.h>
#include <time.h>

#include "tera_types.h"
#include "tera_log.h"
#include "tera_mem.h"
#include "tera_errors.h"
#include "tera_pcoip_link.h"

// This constant determines the path and name of the file containing
// the configuration settings. If this file does not exist the daemon
// uses the default settings defined within the code.
#define DAEMON_CONFIG_FILENAME  "/etc/pcoiphostswd.conf"

// This constant determines the path and name of the lock file used by the
// daemon to ensure only one instance of the daemon is running.
#define DAEMON_LOCK_FILENAME    "/var/run/pcoiphostswd.pid"

// This constant determines how often the daemon checks if the host SW is
// running.
#define DAEMON_LOOP_TIME_SECS   1

#define EXIT_SUCCESS 0
#define EXIT_FAILURE 1

// This variable controls the logging level. Some log messages are generated
// before the logging module is initialized. To view Debug messages in the
// pre-logging init code change the initial value assigned to this variable
// to TERA_LOG_DEBUG and rebuild/reinstall the daemon.
#if 1
eTERA_LOG_LEVEL m_log_level = TERA_LOG_INFO;
#else
eTERA_LOG_LEVEL m_log_level = TERA_LOG_DEBUG;
#endif

// These constants determine the length of character arrays used by the daemon.
#define MAX_LEN_FILENAME        400
#define MAX_LINE_LEN            256

// --------------------------------------------------------------------------
//  eCFG_DATA_TYPE
///     Configuration data types
//
//  Description:
///     This enumeration defines types of configuration data. The code that
///     parses the data in the configuration file uses this type to determine
///     how the data should be interpretted.
///
typedef enum
{
    CFG_TYPE_STR     = 0,
    CFG_TYPE_INT     = 1,
    CFG_TYPE_MAX
} eCFG_DATA_TYPE;

// --------------------------------------------------------------------------
//  sCONFIG_LOOKUP_TABLE_ENTRY
///     Daemon configuration data field entry
//
//  Description:
///     This structure defines the fields in each entry of the configuration
///     data lookup table. The lookup table must contain an entry for each
///     setting found in the configuration file.
///
///     name:    this string is the name of the field in the configuration file
///     type:    this field determines how the parser interprets the data
///              assigned to the field
///     str_val: if the field is a string this field contains a copy of the
///              string
///     int_val: if the field is a integer this field contains a copy of the
///              integer
///
typedef struct sCONFIG_LOOKUP_TABLE_ENTRY_STRUCT
{
    const char           *name;
    const eCFG_DATA_TYPE  type;
    char                  str_val[MAX_LINE_LEN];
    int                   int_val;
} sCONFIG_LOOKUP_TABLE_ENTRY;

#define UNUSED_STR                      ""
#define UNUSED_INT                      0

// The following constants define the defaut values for each configuration
// setting. When new settings are added a default value for the setting should
// be defined here.
#define DAEMON_ENABLE_DEF           1
#define DAEMON_LOGFILE_DEF          "/var/log/pcoiphostswd.log"
#define DAEMON_LOGLEVEL_DEF         TERA_LOG_INFO
#define DAEMON_START_DELAY_DEF      1

//
// The following lookup table contains entries for each supported configuration
// setting.
//
// Notes:
//  1) New settings should be added to the bottom of the table.
//  2) When adding a new configuration field (string or integer):
//     - Set the name entry equal to the string assinged to the new field. This
//       string must match the string found in the configuation file.
//     - Define a new macro for the default value of the field. This value will
//       be used if the configuration file does not include an entry for the
//       field.
//     - add a new entry to the sTERA_HOSTSW_DAEMON_CONFIG structure
//     - update the code near the end of the get_config() function which
//       extracts configuration value from the lookup table and stores it in
//       the sTERA_HOSTSW_DAEMON_CONFIG structure
//  3) When adding a new string configuration field:
//     - set the type field equal to CFG_TYPE_STR
//     - set the str_val field equal to newly defined default macro
//     - set the int_val field equal to UNUSED_INT
//  4) When adding a new integer configuration field:
//     - set the type field equal to CFG_TYPE_INT
//     - set the str_val field equal to UNUSED_STR
//     - set the int_val field equal to newly defined default macro
//
static sCONFIG_LOOKUP_TABLE_ENTRY config_lookup_table[] =
{
    {"DaemonEnable",      CFG_TYPE_INT, UNUSED_STR,               DAEMON_ENABLE_DEF},
    {"DaemonLogFile",     CFG_TYPE_STR, DAEMON_LOGFILE_DEF,       UNUSED_INT},
    {"DaemonLogLevel",    CFG_TYPE_INT, UNUSED_STR,               DAEMON_LOGLEVEL_DEF},
    {"DaemonStartDelay",  CFG_TYPE_INT, UNUSED_STR,               DAEMON_START_DELAY_DEF},
};

#define LOOKUP_TABLE_ENTRIES    ((int)(sizeof(config_lookup_table)/sizeof(sCONFIG_LOOKUP_TABLE_ENTRY)))

// --------------------------------------------------------------------------
//  sTERA_HOSTSW_DAEMON_CONFIG
///     Daemon configuration data
//
//  Description:
///     This structure defines the configuration data settings that control
///     how the daemon operates.
///
typedef struct sTERA_HOSTSW_DAEMON_CONFIG_STRUCT
{
    int             daemon_enable;                   ///< enable/disable daemon
    char            logfile[MAX_LEN_FILENAME];       ///< daemon log filename
    eTERA_LOG_LEVEL log_level;                       ///< daemon logging verbosity
    int             daemon_start_delay;              ///< # seconds daemon waits
                                                     ///< before entering the loop
                                                     ///< that check if the host SW
                                                     ///< is running
} sTERA_HOSTSW_DAEMON_CONFIG;

// --------------------------------------------------------------------------
//  get_field_name_index_and_type
///     Convert field name into an index
//
//  Description:
///     This function converts a configuration file field name into an index
///     into the lookup table. If the field name is not found in the lookup
///     table the function returns -1, otherwise the function return the index
///     of the lookup table entry.
//
static int get_field_name_index(
    const char *field_name)
{
    int i;

    for (i=0; i<LOOKUP_TABLE_ENTRIES; i++)
    {
        if (!strcmp(field_name, config_lookup_table[i].name))
        {
            mTERA_LOG_DBG("matched field '%s' at index %d", field_name, i);
            return i;
        }
    }
    mTERA_LOG_ERR("failed to find '%s'", field_name);
    return -1;
} // get_field_name_index()

// --------------------------------------------------------------------------
//  get_config
///     Dump configuration data
//
//  Description:
///     Print the value of the configuration settings.
//
static void dump_config(const sTERA_HOSTSW_DAEMON_CONFIG *cfg)
{
    mTERA_LOG_INFO("logfile %s level %d daemon_en %d Daemon dly %d",
                   cfg->logfile,
                   cfg->log_level,
                   cfg->daemon_enable,
                   cfg->daemon_start_delay);
} // dump_config()

// --------------------------------------------------------------------------
//  get_config
///     Get configuration data
//
//  Description:
///     This function updates the configuration data structure using data
///     in the lookup table. The lookup table is statically initialized with
///     default values for each configuration setting. This function
///     overwrites the default values in the lookup table with settings
///     found in the configuration file. After processing the configuration
///     file the values in the lookup table are copied to the configuration
///     data structure pointer passed into this function by the caller.
//
static void get_config(sTERA_HOSTSW_DAEMON_CONFIG *cfg)
{
    FILE           *fp;
    char            line[MAX_LINE_LEN];
    char           *field_name;
    char           *field_value;
    int             field_idx;

    // open the config file
    fp = fopen(DAEMON_CONFIG_FILENAME, "r");
    if (!fp)
    {
        mTERA_LOG_INFO("Unable to open daemon config file %s", DAEMON_CONFIG_FILENAME);
    }
    else
    {
        // read and process each line of the configuration file and overwrite
        // the default values in the lookup table
        while (fgets(line, MAX_LINE_LEN, fp) != NULL)
        {
            // skip comment lines
            if (line[0] == '#')
            {
                continue;
            }

            // read the field name
            field_name = strtok(line, " =\n");
            if (field_name != NULL)
            {
                mTERA_LOG_DBG("Processing field name: %s", field_name);

                // convert field name to an index in the lookup table
                field_idx = get_field_name_index(field_name);

                // skip line if index is not valid
                if ((field_idx < 0) || (field_idx >= LOOKUP_TABLE_ENTRIES))
                {
                    mTERA_LOG_DBG("returned idx %d, skipping line: %s", field_idx, line);
                    continue;
                }

                // Read value of field from config file. Note the space
                // character has been removed from the list of tokens. This
                // allows users to specify strings, which is useful when
                // including options with the string that starts the host SW.
                field_value = strtok(NULL, "=\n");
                mTERA_LOG_DBG("value = %s", field_value);

                // update field value in lookup table
                switch (config_lookup_table[field_idx].type)
                {
                    case CFG_TYPE_STR:
                        strncpy(config_lookup_table[field_idx].str_val, field_value, MAX_LINE_LEN);
                        break;
                    case CFG_TYPE_INT:
                        config_lookup_table[field_idx].int_val = atoi(field_value);
                        break;
                    default:
                        mTERA_LOG_ERR("Illegal type %d in lookup table at index %d field %s",
                                      config_lookup_table[field_idx].type,
                                      field_idx,
                                      field_name);
                        break;
                }
            }
        }
    }

    // copy lookup table values to the config data structure
    cfg->daemon_enable       = config_lookup_table[0].int_val;
    strncpy(cfg->logfile,       config_lookup_table[1].str_val, MAX_LEN_FILENAME);
    cfg->log_level           = (eTERA_LOG_LEVEL)config_lookup_table[2].int_val;
    cfg->daemon_start_delay  = config_lookup_table[3].int_val;

} // get_config()

// --------------------------------------------------------------------------
//  init_logging
///     Initialize logging
//
//  Description:
///     This function initializes the logging module. It updates the global
///     logging level variable that controls what log messages (debug, info,
///     error and critical) are generated by the daemon. It opens the log file
///     and redirects STDOUT and STDERR to this file. The logging macros write
///     log messages to STDOUT.
//
static int init_logging(
    eTERA_LOG_LEVEL  level,
    char            *filename)
{
    // initialize the global logging level
    if (level < TERA_LOG_MAX)
    {
        mSET_LOG_LEVEL(level);
    }
    else
    {
        mSET_LOG_LEVEL(TERA_LOG_ERROR);
    }

    // if any log messages are enabled
    if (m_log_level != TERA_LOG_NONE)
    {
        // open the log file giving all users read/write access
        int fd = open(filename, O_APPEND|O_CREAT|O_WRONLY, 0666);
        if (fd == -1)
        {
            mTERA_LOG_ERR("Error opening logfile %s (ret %d)", filename, fd);
            return -1;
        }

        // redirect stdout and stderr to the logfile and close the open file
        dup2(fd, STDOUT_FILENO);
        dup2(fd, STDERR_FILENO);
        close(fd);
    }

    mTERA_LOG_INFO("success: level %d, file %s", level, filename);
    return 0;
} // init_logging()

// --------------------------------------------------------------------------
//  daemon_handler
///     Daemon signal handler
//
//  Description:
///     This function is the signal handler for the daemon.
//
static void daemon_handler(int signum)
{
    mTERA_LOG_INFO("daemon_handler(): signum=%d pid=%d", signum, getpid());
    switch(signum)
    {
        case SIGALRM: exit(EXIT_FAILURE); break;
// added 3/27
        case SIGHUP:  break;
        case SIGUSR1: exit(EXIT_SUCCESS); break;
        case SIGCHLD: exit(EXIT_FAILURE); break;
        case SIGTERM: exit(EXIT_FAILURE); break;
        default: exit(EXIT_FAILURE); break;
    }
} // daemon_handler()

// --------------------------------------------------------------------------
//  child_handler
///     Child process (host SW) signal handler
//
//  Description:
///     This function is the signal handler for the process the host SW
///     runs under. The difference between this handler and the daemon handler
///     is the fact that this function calls "_exit()" instead of "exit()".
///     X IO error messages are generated if the exit() function is called by
///     the process that previously ran the host SW. Details about this are
///     on https://developer.gnome.org/gtk-faq/stable/x505.html.
//
static void child_handler(int signum)
{
    mTERA_LOG_INFO("child_handler(): signum=%d pid=%d", signum, getpid());
    switch(signum)
    {
        case SIGALRM: _exit(EXIT_FAILURE); break;
        case SIGHUP:  _exit(EXIT_FAILURE); break;
        case SIGUSR1: _exit(EXIT_SUCCESS); break;
        case SIGCHLD: _exit(EXIT_FAILURE); break;
        case SIGTERM: _exit(EXIT_FAILURE); break;
        default: _exit(EXIT_FAILURE); break;
    }
} // child_handler()

// --------------------------------------------------------------------------
//  daemonize
///     Start the daemon
//
//  Description:
///     This function starts the host SW daemon process. The function calls
///     fork to create a child process which the daemon will run under and the
///     parent process will exit.
///
///     The child process ensures only one instance of the daemon is running
///     by using a lock file. If another instance of the daemon is running the
///     child process exits.
//
static void daemonize(void)
{
    pid_t pid, sid, parent;
    int lock_fp = -1;
    char pid_str[MAX_LINE_LEN];

    // already a daemon
    if (getppid() == 1) return;

    // Trap signals that we expect to recieve
    signal(SIGCHLD,daemon_handler);
    signal(SIGUSR1,daemon_handler);
    signal(SIGALRM,daemon_handler);
    signal(SIGTERM,daemon_handler);
    signal(SIGHUP, daemon_handler);

    // Fork off the parent process
    pid = fork();
    if (pid < 0)
    {
        mTERA_LOG_ERR("Unable to fork daemon, code=%d (%s)", errno, strerror(errno));
        exit(EXIT_FAILURE);
    }

    // If we got a good PID, then we can exit the parent process
    if (pid > 0)
    {
        // Wait for confirmation from the child via SIGTERM or SIGCHLD, or
        // for two seconds to elapse (SIGALRM).  pause() should not return.
        alarm(2);
        pause();

        exit(EXIT_SUCCESS);
    }

    // At this point we are executing as the child process
    parent = getppid();

    // Create a lock file. If this fails the daemon exits because another
    // instance of the daemon is running. The file will remain locked until
    // the daemon exits.
    lock_fp = open(DAEMON_LOCK_FILENAME, O_RDWR | O_CREAT, 0640);
    if (lock_fp == -1)
    {
        mTERA_LOG_ERR("Unable to open lockfile %s, code %d (%s)",
                      DAEMON_LOCK_FILENAME, errno, strerror(errno));
        exit(EXIT_FAILURE);
    }
    if (lockf(lock_fp, F_TLOCK, 0) == -1)
    {
        mTERA_LOG_ERR("Unable to lock lockfile %s, code %d (%s)",
                      DAEMON_LOCK_FILENAME, errno, strerror(errno));
        exit(EXIT_FAILURE);
    }

    // store the process PID in the lock file
    snprintf(pid_str, sizeof(pid_str), "%d\n", getpid());
    write(lock_fp, pid_str, strlen(pid_str));

    // Cancel certain signals
    signal(SIGCHLD,SIG_DFL); // A child process dies
    signal(SIGTSTP,SIG_IGN); // Various TTY signals
    signal(SIGTTOU,SIG_IGN);
    signal(SIGTTIN,SIG_IGN);
    signal(SIGHUP, SIG_IGN); // Ignore hangup signal
    signal(SIGTERM,SIG_DFL); // Die on SIGTERM

    // Change the file mode mask
    umask(0);

    // Create a new SID for the child process
    sid = setsid();
    if (sid < 0)
    {
        mTERA_LOG_ERR("Unable to create a new session, code %d (%s)",
                      errno, strerror(errno));
        exit(EXIT_FAILURE);
    }

    // Change the current working directory. This prevents the current
    // directory from being locked; hence not being able to remove it.
    if ((chdir("/")) < 0)
    {
        mTERA_LOG_ERR("Unable to change directory to %s, code %d (%s)",
                      "/", errno, strerror(errno));
        exit(EXIT_FAILURE);
    }

    // Redirect standard input to /dev/null
    freopen("/dev/null", "r", stdin);

    // if logging is not enabled redirect standard output to /dev/null
    if (m_log_level == TERA_LOG_NONE)
    {
        freopen("/dev/null", "w", stdout);
        freopen("/dev/null", "w", stderr);
    }

    // notify parent process the daemon is running
    kill(parent, SIGUSR1);
    mTERA_LOG_INFO("sent signal %d to parent process", SIGUSR1);
} // daemonize()

//
// TBD - This code may be used if the daemon needs to kill the host SW when a
//       user logs out. Currently all testing shows the host SW stops when a
//       user logs out.
//
#if 0
// --------------------------------------------------------------------------
//  is_hostsw_active
///     Check if the host SW is running
//
//  Description:
///     This function returns zero if the Host SW executable (pcoip_agent) is
///     not running and 1 if it is running. If an error occurs and the function
///     cannot determine if the host SW is running the function returns -1.
//
static int is_hostsw_active(void)
{
    FILE *fp;
    int ret;
    int rv=-1;
    uid_t uid_hostsw;
    int pid_hostsw;

    fp = popen("ps -o pid= -o uid= -C pcoip_agent", "r");
    if (fp == NULL)
    {
        mTERA_LOG_ERR("popen() failed, code=%d (%s)", errno, strerror(errno));
    }
    else
    {
        ret = fscanf(fp, "%d %d", &pid_hostsw, &uid_hostsw);
        if (ret <= 0)
        {
            mTERA_LOG_DBG("pcoip_agent not active");
            rv = 0;
        }
        else
        {
            mTERA_LOG_DBG("pcoip_agent active, hostsw pid = %d uid=%d, daemon pid %d",
                          pid_hostsw, uid_hostsw, getpid());
            rv = 1;
        }

        ret = pclose(fp);

        // Log error message if pclose failed. Note: pclose returns -1
        // and sets errno = ECHILD when the child process created by the
        // popen() call above has finished, which is expected. This is why
        // the code below ignores this error.
        if ((ret == -1) && (errno != ECHILD))
        {
            mTERA_LOG_ERR("pclose() failed, code=%d (%s), pid %d uid %d",
                          errno, strerror(errno), pid_hostsw, uid_hostsw);
        }
    }

    if (rv == -1)
    {
        mTERA_LOG_ERR("Unable to determine whether host SW is running!");
    }
    return rv;
} // is_hostsw_active()
#endif

// --------------------------------------------------------------------------
//  is_user_logged_in_locally
///     Check if a user is logged in locally
//
//  Description:
///     This function returns zero if a user is not logged in locally and 1 if
///     a user is logged in locally. If an error occurs and the function cannot
///     determine if a user is logged in locally the function returns -1.
///
///     The function uses the "who -u" command to determine if a user is logged
///     in locally. This command returns a list of the users logged in.
///     This function considers a user to be logged in locally if the output
///     contains the string "(:" or the second column starts with the colon
///     character ':' followed by a decimal number.
///
///     Below is a sample of the output when the pcoiplab user is logged in
///     locally and the apexlab user is logged in using ssh on a CentOS 6.4
///     system.
///
///       NAME     LINE         TIME             IDLE          PID COMMENT
///       pcoiplab tty7         2013-04-12 16:39  old         7076 (:0)
///       pcoiplab pts/0        2013-04-12 16:39 00:24        7315 (:0.0)
///       pcoiplab pts/1        2013-04-12 16:39 00:10        7315 (:0.0)
///       pcoiplab pts/2        2013-04-12 16:39   .          7315 (:0.0)
///       pcoiplab pts/3        2013-04-12 16:39 00:14        7315 (:0.0)
///       apexlab  pts/4        2013-04-12 16:38 00:06        6805 (10.200.2.10)
///
///     The first 5 entries are associated with the locally logged in user and
///     the last line is associated with the SSH session.
///
///     Below is a sample of the output when the pcoiplab user is logged in
///     locally and the apexlab user is logged in using ssh on a CentOS 5.9
///     system.
///
///       NAME     LINE         TIME             IDLE          PID COMMENT
///       apexlab  pts/1        2013-06-24 14:34   .         31970 (10.200.2.10)
///       pcoiplab :0           2013-06-24 14:36   ?         32123
///       pcoiplab pts/2        2013-06-24 14:50   .          1706 (:0.0)
///
///     The last 2 entries are associated with the locally logged in user and
///     the first line is associated with the SSH session.
//
static int is_user_logged_in_locally(void)
{
    FILE *fp;
    int ret;
    int rv=-1;

    fp = popen("who -u", "r");
    if (fp == NULL)
    {
        mTERA_LOG_ERR("popen() failed, code=%d (%s)", errno, strerror(errno));
    }
    else
    {
        rv = 0;
        char line[500];
        char username[100];
        int  disp_num;
        while(fgets(line,sizeof line,fp)!= NULL) /* read a line from a file */
        {
            mTERA_LOG_DBG("'who -u' data: %s", line);

            // Check for CentOS 5
            if (sscanf(line, "%s :%d", username, &disp_num) == 2)
            {
                // exit if a user is logged in locally
                mTERA_LOG_DBG("user %s logged in locally :%d", username, disp_num);
                rv = 1;
                break;
            }
            // Check for CentOS 6
            else if (strstr(line, " (:") != NULL)
            {
                // exit if a user is logged in locally
                mTERA_LOG_DBG("user logged in locally");
                rv = 1;
                break;
            }
        }

        ret = pclose(fp);

        // Log error message if pclose failed. Note: pclose returns -1
        // and sets errno = ECHILD when the child process created by the
        // popen() call above has finished, which is expected. This is why
        // the code below ignores this error.
        if ((ret == -1) && (errno != ECHILD))
        {
            mTERA_LOG_ERR("pclose() failed, code=%d (%s)", errno, strerror(errno));
        }
    }

    if (rv == -1)
    {
        mTERA_LOG_ERR("Unable to determine whether user logged in locally!");
    }
    return rv;
} // is_user_logged_in_locally()

// --------------------------------------------------------------------------
//  start_hostsw
///     Create a child process that runs the host SW
//
//  Description:
///     This function uses the fork() function to create a child process that
///     starts the host SW. The host SW is started using a script. If the
///     host SW returns the child process exits.
//
int start_hostsw(
    const char *script_name)
{
    pid_t   pid_fork;
    int     ret;

    // create child process that the Host SW will run in
    pid_fork = fork();

    // if this is the child process prepare and start host SW
    if (pid_fork == 0)
    {
        mTERA_LOG_INFO("host SW process created, starting Host SW with '%s'",
                       script_name);

        // configure this process to exit when a user logs in or out
        signal(SIGHUP,child_handler);
        signal(SIGUSR1,child_handler);
        signal(SIGALRM,child_handler);
        signal(SIGTERM,child_handler);


        // start the host SW
        ret = system(script_name);   // system doc says do not use the functon with set-user-ID

        mTERA_LOG_INFO("Host SW done, occurs when logging in or host SW exits (ret %d)",
                       ret);

        // exit the child process running the host SW when the host SW exits
        _exit(1);
    }
    else if (pid_fork < 0)
    {
        // fork failed
        mTERA_LOG_ERR("fork() failed, code=%d (%s), exiting!",
                      errno, strerror(errno));
        return 0;
    }
    else
    {
        // parent process
        mTERA_LOG_INFO("fork successfully created host SW process (pid %d)", pid_fork);
    }
    return 1;
} // start_hostsw()

// --------------------------------------------------------------------------
//  update_driver_cfg
///     Update the driver configuration settings
//
//  Description:
///     This function updates the driver configuation settings.
//
int update_driver_cfg(
    cTERA_PCOIP_LINK                   *pcoip_link,
    eTERA_PCOIP_LINK_LCUR_EVT_ALG_TYPE  alg)
{
    int ret=1;

    if (pcoip_link->kmp_driver_cfg_set(alg) != TERA_SUCCESS)
    {
        mTERA_LOG_ERR("failed to configure cursor mode %d in driver", alg);
        ret = 0;
    }

    return ret;
} // update_driver_cfg()

// --------------------------------------------------------------------------
//  enable_local_cursor_with_relative_reports
///     Enable local cursor feature with relative HID reports
//
//  Description:
///     This function enables the local cursor feature and configures the
///     driver to submit HID reports as relative data.
//
int enable_local_cursor_with_relative_reports(
    cTERA_PCOIP_LINK                   *pcoip_link)
{
    int ret;
    int retry=20;

    while (retry > 0)
    {
        // initialize the return value each time through the loop
        ret=1;

        if (pcoip_link->hdc_open() == TERA_SUCCESS)
        {
            // Wait a couple seconds to allow the system to stabilize.
            // If a session just started it will take some time to
            // process the monitor hot plug signal(s).
            sleep(2);

            if (pcoip_link->lcur_lkbd_enable() == TERA_SUCCESS)
            {
                if (pcoip_link->lcur_pointer_visibility_set(0,0) != TERA_SUCCESS)
                {
                    mTERA_LOG_ERR("failed to hide local cursor");
                    ret = 0;
                }

                // exit the loop if the driver is updated successfully and no
                // errors occurred
                if ((update_driver_cfg(pcoip_link, TERA_PCOIP_LINK_LCUR_REL_ALG)) &&
                    (ret == 1))
                {
                    retry = 0;
                }
            }
            else
            {
                mTERA_LOG_ERR("lcur_lkbd_enable() failed");
                ret = 0;
            }
        }
        else
        {
            mTERA_LOG_ERR("hdc_open() failed");
            ret = 0;
        }
        retry--;
    }

    // if an error occurred make sure driver uses the absolute algorithm
    if (ret == 0)
    {
        update_driver_cfg(pcoip_link, TERA_PCOIP_LINK_LCUR_ABS_ALG);
    }
    return ret;
} // enable_local_cursor_with_relative_reports()

// --------------------------------------------------------------------------
//  disable_local_cursor
//      Disable the local cursor and keyboard
//
//  Description:
//      This function sends a message to disable the local cursor and keyboard.
//
void disable_local_cursor(
    cTERA_PCOIP_LINK  *pcoip_link)
{

    // Blindly disable local cursor and keyboard
    pcoip_link->lcur_lkbd_disable();

} // disable_local_cursor()

// --------------------------------------------------------------------------
//  main
///     Main function of the daemon
//
//  Description:
///     This function creates and controls a daemon process. The daemon's
///     job is to enable the local cursor feature for SC sessions when the
///     login screen is active.
//
int main( int argc, char *argv[] )
{
    sTERA_HOSTSW_DAEMON_CONFIG             config_data;
    cTERA_PCOIP_LINK                       pcoip_link;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    INT                                    sess_link_change;
    int                                    login_state_change;
    int                                    user_logged_in;

    // verify the daemon is being run by root
    if (getuid() != 0)
    {
        mTERA_LOG_ERR("The daemon must be run by root, exiting!");
        return 0;
    }

    // Read the configuration settings
    get_config(&config_data);

    // Exit if the daemon is disabled
    if (!config_data.daemon_enable)
    {
        mTERA_LOG_INFO("The pcoiphostswd daemon is disabled");
        return 0;
    }

    // Initialize the logging interface
    if (init_logging(config_data.log_level, config_data.logfile) == -1)
    {
        mTERA_LOG_ERR("init_logging() failed, exiting!");
        return 0;
    }
    dump_config(&config_data);

    // Daemonize
    daemonize();
    mTERA_LOG_INFO("pcoiphostswd daemon running!");

    // ignore signal from children, if this is not done the forked processes
    // enter a zombie state when the hostsw exits
    signal(SIGCHLD, SIG_IGN);

    // update variable used to determine when a user logs in or logs out
    user_logged_in = is_user_logged_in_locally();

    // delay the start of the daemon to allow other components in the system
    // to complete initialization, this only affects the daemon when it starts
    sleep(config_data.daemon_start_delay);

    // Start the PCoIP link object
    pcoip_link.start();

    // do the daemon work
    while (1)
    {
        // Determine what changed
        login_state_change = 0;
        sess_link_change = 0;
        pcoip_link.link_heartbeat(&sess_link_change,
                                  &sess_link);
        if (is_user_logged_in_locally() != user_logged_in)
        {
            login_state_change = 1;
            user_logged_in = !user_logged_in;
        }

        // If local user logged out and a session is active
        if ((login_state_change) &&
            (!user_logged_in) &&
            (sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
        {
            // Enable local cursor if a SC session active
            if (pcoip_link.is_client_sc())
            {
                mTERA_LOG_DBG("Enable local cursor in relative mode: SC logout while session active");
                enable_local_cursor_with_relative_reports(&pcoip_link);
            }
            else
            {
                mTERA_LOG_DBG("Disable local cursor: ZC logout while session active");
                disable_local_cursor(&pcoip_link);
            }
        }
        // Enable local cursor if login screen is active and SC session started
        else if ((!user_logged_in) &&
                 (sess_link_change) &&
                 (sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED) &&
                 (pcoip_link.is_client_sc()))
        {
            mTERA_LOG_DBG("Enable local cursor in relative mode: session start while login screen active");
            enable_local_cursor_with_relative_reports(&pcoip_link);
        }

        // periodically sleep
        sleep(DAEMON_LOOP_TIME_SECS);
    }

    // cleanup
    update_driver_cfg(&pcoip_link, TERA_PCOIP_LINK_LCUR_ABS_ALG);
    pcoip_link.stop();

    mTERA_LOG_INFO("terminated");
    return 0;
} // main()
