////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Features Property Page Class Implementation
///
///     This file implements the PCoIP Host Software Features property page
///     class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include <wx/tooltip.h>
#include <wx/cursor.h>

#include "tera_log.h"
#include "tera_mem.h"
#include "tera_types.h"
#include "tera_errors.h"
#include "tera_pcoip_agent_gui_prop_feat.h"
#include "tera_pcoip_agent_internal.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// Local cursor shape visibility table
static const wxString pcoip_agent_cur_visibile_str_tbl[] =
{
    _T("Local and Remote"),
    _T("Remote"),
    _T("Local"),
};


BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI_PROP_FEAT, wxPanel)
    EVT_BUTTON(wxID_APPLY, cPCOIP_AGENT_GUI_PROP_FEAT::evt_apply_btn)
    EVT_CHECKBOX(wxID_LCUR_LKBD_EN_CHK,
                 cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_lkbd_en_chk)
    EVT_CHECKBOX(wxID_CHECK_LKBD_REPEAT,
                 cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_lkbd_en_chk)
    EVT_CHOICE(wxID_LCUR_VIS_CHOICE,
               cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_vis_choice)

    // Catch-all for any commands that do not require special processing
    EVT_CHECKBOX(wxID_ANY, cPCOIP_AGENT_GUI_PROP_FEAT::evt_any)
    EVT_CHOICE(wxID_ANY, cPCOIP_AGENT_GUI_PROP_FEAT::evt_any)
    EVT_COMMAND_SCROLL(wxID_ANY, cPCOIP_AGENT_GUI_PROP_FEAT::evt_any_scroll)

    EVT_INIT_DIALOG(cPCOIP_AGENT_GUI_PROP_FEAT::evt_init_dialog)
    EVT_CLOSE(cPCOIP_AGENT_GUI_PROP_FEAT::evt_close)
END_EVENT_TABLE()

//
// the following constants are used to scale the keyboard repeat rate and delay
// slider settings to be compatible with the settings used by the client device
//
#define LKBD_DELAY_NUM_TICS     5
#define LKBD_RATE_NUM_TICS      11
#define LKBD_DELAY_RANGE        (TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DELAY_MAX - \
                                  TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DELAY_MIN)
#define LKBD_RATE_RANGE         (TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_RATE_MAX - \
                                 TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_RATE_MIN)
#define LKBD_DELAY_INC_PER_TIC  (LKBD_DELAY_RANGE / (LKBD_DELAY_NUM_TICS - 1))
#define LKBD_RATE_INC_PER_TIC   (LKBD_RATE_RANGE  / (LKBD_RATE_NUM_TICS - 1))

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_apply_btn(
    wxCommandEvent& evt
    )
{
    bool                                   lock_ws;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    UINT32                                 lcur_spd;
    UINT8                                  lcur_lkbd_en;
    UINT32                                 lcur_vis;
    UINT8                                  lcur_fade;

    // write the feature settings that apply to all sessions
    lock_ws = m_lock_host_pc_en_chk->IsChecked();
    wxGetApp().gui().host_cfg().set_ws_lock(lock_ws);

    // get the MAC address of the client
    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    if (m_lcur_lkbd_en_chk->IsEnabled())
    {
        lcur_lkbd_en = m_lcur_lkbd_en_chk->GetValue();
        mTERA_LOG_DBG("local cursor/keyboard enable '%d'", lcur_lkbd_en);

        // Store the local cursor/keyboard setting into the user profile
        wxGetApp().gui().host_cfg().set_lcur_lkbd_en_setting(sess_link.mac,
                                                             (int)lcur_lkbd_en);
    }

    if (m_lcur_vis_choice->IsEnabled())
    {
        lcur_vis = m_lcur_vis_choice->GetCurrentSelection();
        wxGetApp().gui().host_cfg().set_lcur_vis_setting(sess_link.mac,
                                                         (int)lcur_vis);

        // if the client supports hiding the cursor AND the remote cursor
        // is visible enable cursor fade option
        if ((wxGetApp().gui().pcoip_link().lcur_idle_hide_avail() == TERA_SUCCESS) &&
            (lcur_vis != PCOIP_AGENT_CUR_VISIBLE_HOST))
        {
            m_lcur_fade_chk->Enable();
        }
        else
        {
            m_lcur_fade_chk->Disable();
        }
    }

    // Configure the client keyboard settings
    if (m_check_lkbd_repeat->IsEnabled())
    {
        // Store client keyboard repeat enable setting in the user profile
        wxGetApp().gui().host_cfg().set_repeat_en_setting(sess_link.mac,
                                                          (int)m_check_lkbd_repeat->GetValue());

        // Store client keyboard repeat settings if the feature is enabled
        if (m_check_lkbd_repeat->GetValue())
        {
            // always write the settings to the client
            mTERA_LOG_DBG("writing kbd repeat settings to client");
            wxGetApp().gui().pcoip_link().lkbd_repeat_rates_set(lkbd_delay_get_gui(),
                                                                lkbd_rate_get_gui());
            //
            // only write the settings to the registry if the client does not
            // support reading the settings
            //
            if (m_client_kbd_setting_avail != TERA_SUCCESS)
            {
                mTERA_LOG_DBG("writing kbd repeat settings to registry");
                wxGetApp().gui().host_cfg().set_repeat_delay_setting(sess_link.mac,
                                                                     (int)lkbd_delay_get_gui());
                wxGetApp().gui().host_cfg().set_repeat_rate_setting(sess_link.mac,
                                                                    (int)lkbd_rate_get_gui());
            }
        }
    }

    if (m_lcur_fade_chk->IsEnabled())
    {
        lcur_fade = m_lcur_fade_chk->GetValue();
        wxGetApp().gui().host_cfg().set_lcur_fade_setting(sess_link.mac,
                                                          (int)lcur_fade);
    }

    // send the latest settings to the client
    wxGetApp().gui().lcur_lkbd_settings_send(wxGetApp().gui().display_topology_ok());

    // All changes are applied - disable the Apply button
    m_apply_btn->Disable();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_apply_btn()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_lkbd_en_chk(
    wxCommandEvent& evt
    )
{
    lcur_lkbd_ctrl_update();

    m_apply_btn->Enable();
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_lkbd_en_chk()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_vis_choice(
    wxCommandEvent& evt
    )
{
    m_apply_btn->Enable();
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_lcur_vis_choice()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_any(
    wxCommandEvent& evt
    )
{
    m_apply_btn->Enable();
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_any()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_any_scroll(
    wxScrollEvent& evt
    )
{
    m_apply_btn->Enable();
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_any_scroll()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_init_dialog(
    wxInitDialogEvent& evt
    )
{
    mTERA_LOG_DBG("FEAT EVT_INIT_DIALOG");

    // Initialize the security controls
    init_dialog_security();

    // Initialize the local cursor/keyboard controls
    init_dialog_lcur_lkbd();

    // Disable the Apply button
    m_apply_btn->Disable();

    evt.Skip();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_init_dialog()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_close(
    wxCloseEvent& evt
    )
{
    mTERA_LOG_DBG("start - cPCOIP_AGENT_GUI_PROP_FEAT");

    evt.Skip();

    mTERA_LOG_DBG("end - cPCOIP_AGENT_GUI_PROP_FEAT");
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_close()

void cPCOIP_AGENT_GUI_PROP_FEAT::evt_update_ui(
    wxUpdateUIEvent& evt
    )
{
    mTERA_LOG_DBG("UPDATE UI!!!!!");
}

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::init_dialog_security
//      Initialize the security part of the page
//
//  Description:
//      This function initializes the security part of the
//      property page.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::init_dialog_security(void)
{

    // Set the enable box according to the user's preference
    m_lock_host_pc_en_chk->SetValue(wxGetApp().gui().host_cfg().get_ws_lock());

} // cPCOIP_AGENT_GUI_PROP_FEAT::init_dialog_security()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::init_lcur_slider
//      Initialize the local cursor speed slider
//
//  Description:
//      This function initializes the local cursor speed slider.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::init_lcur_slider(bool enable)
{
    // Initialize local cursor speed control
    m_lcur_spd_slider->Enable(enable);
    m_lcur_spd_txt->Enable(enable);
    m_lcur_spd_slow_txt->Enable(enable);
    m_lcur_spd_fast_txt->Enable(enable);

} // cPCOIP_AGENT_GUI_PROP_FEAT::init_lcur_slider()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::init_lkbd_sliders
//      Initialize the local keyboard repeat sliders
//
//  Description:
//      This function initializes the local keyboard repeat sliders (Delay
//      and Rate).
//
void cPCOIP_AGENT_GUI_PROP_FEAT::init_lkbd_sliders(bool enable)
{
    // Initialize local keyboard repeat delay controls
    m_slider_lkbd_delay->Enable(enable);
    m_lkbd_delay_txt->Enable(enable);
    m_lkbd_delay_long_txt->Enable(enable);
    m_lkbd_delay_short_txt->Enable(enable);

    // Initialize local keyboard repeat rate controls
    m_slider_lkbd_rate->Enable(enable);
    m_lkbd_rate_txt->Enable(enable);
    m_lkbd_rate_slow_txt->Enable(enable);
    m_lkbd_rate_fast_txt->Enable(enable);

} // cPCOIP_AGENT_GUI_PROP_FEAT::init_lkbd_sliders()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::init_dialog_lcur_lkbd
//      Initialize the local cursor/keyboard part of the page
//
//  Description:
//      This function initializes the local cursor/keyboard part of the
//      property page.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::init_dialog_lcur_lkbd(void)
{
    UINT32                                 lcur_vis;
    UINT32                                 lcur_spd;
    UINT32                                 stat;
    UINT32                                 cnt;
    INT32                                  ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    topo_ok;
    int                                    screen_cfg_ok;
    int                                    lkbd_repeat_en_cfg;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    // Set the enable box according to the user's preference
    m_lcur_lkbd_en_chk->SetValue(wxGetApp().gui().lcur_lkbd_enable_get());

    // Set the fadeout setting with value from the user's profile
    m_lcur_fade_chk->SetValue(wxGetApp().gui().lcur_fade_get());

    // Set the client keyboard repeat setting with value from the user's profile
    wxGetApp().gui().host_cfg().get_repeat_en_setting(sess_link.mac,
                                                      &lkbd_repeat_en_cfg);
    m_check_lkbd_repeat->SetValue(lkbd_repeat_en_cfg);

    // Check if local cursor and keyboard is available
    topo_ok = wxGetApp().gui().display_topology_ok();
    screen_cfg_ok = wxGetApp().gui().m_prop_sheet.m_mon_page->is_screen_cfg_ok();
    if ((wxGetApp().gui().pcoip_link().lcur_lkbd_avail() != TERA_SUCCESS) ||
        (!topo_ok) ||
        (!screen_cfg_ok))
    {
        mTERA_LOG_DBG("local cursor and keyboard not available: topo_ok=%d screen_cfg_ok=%d",
                      topo_ok, screen_cfg_ok);

        // Local cursor and keyboard not available so hide all related options
        m_lcur_lkbd_en_chk->Enable(false);

        m_lcur_vis_choice->Enable(false);
        m_lcur_vis_txt->Enable(false);

        init_lcur_slider(false);

        m_check_lkbd_repeat->Enable(false);
        init_lkbd_sliders(false);

        m_lcur_fade_chk->Enable(false);

        // Set up and show the local cursor/keyboard help button and show it
        lcur_hlp_btn_setup(1);
    }
    else
    {
        mTERA_LOG_DBG("local cursor and keyboard available");

        lcur_hlp_btn_setup(0);
        //
        // Do not allow changes to the local cursor feature if it cannot be
        // disabled
        //
        if(wxGetApp().gui().pcoip_link().lcur_disable_avail() != TERA_SUCCESS)
        {
            m_lcur_lkbd_en_chk->Enable(false);
        }
        else
        {
            m_lcur_lkbd_en_chk->Enable(true);
        }

        //
        // Cursor visibility and fadeout controls
        //
        if (wxGetApp().gui().pcoip_link().lcur_shape_bmp_send_avail() != TERA_SUCCESS)
        {
            mTERA_LOG_DBG("cannot send local cursor shapes");

            // Disable related controls and hide the cursor preview
            m_lcur_vis_choice->Enable(false);
            m_lcur_vis_txt->Enable(false);

            m_lcur_fade_chk->Enable(false);
        }
        else
        {
            m_lcur_vis_choice->Enable(true);
            m_lcur_vis_txt->Enable(true);

            // First clear all existing items
            m_lcur_vis_choice->Clear();

            // Populate the drop down list with visible cursor options
            for (UINT8 i = 0; i < PCOIP_AGENT_CUR_VISIBLE_NUM; i++)
            {
                // only populate the CLIENT (local) option if the system
                // supports hiding the host cursor
                if (i == PCOIP_AGENT_CUR_VISIBLE_CLIENT)
                {
                    if (wxGetApp().gui().is_host_cur_hiding_supported())
                    {
                        m_lcur_vis_choice->Append(pcoip_agent_cur_visibile_str_tbl[i]);
                    }
                }
                else
                {
                    m_lcur_vis_choice->Append(pcoip_agent_cur_visibile_str_tbl[i]);
                }
            }

            // Get the last used visibility setting from the registry
            lcur_vis = cur_visibility_get_reg();
            m_lcur_vis_choice->SetSelection(lcur_vis);

            // if the client supports hiding the cursor AND the remote cursor
            // is visible enable cursor fade option
            if ((wxGetApp().gui().pcoip_link().lcur_idle_hide_avail() == TERA_SUCCESS) &&
                (lcur_vis != PCOIP_AGENT_CUR_VISIBLE_HOST))
            {
                m_lcur_fade_chk->Enable();
            }
            else
            {
                m_lcur_fade_chk->Disable();
            }
        }

        //
        // Local keyboard repeat controls
        //
        if (wxGetApp().gui().pcoip_link().lkbd_client_auto_repeat_avail() != TERA_SUCCESS)
        {
            mTERA_LOG_DBG("local keyboard repeat not available");

            m_check_lkbd_repeat->Enable(false);

            init_lkbd_sliders(false);
        }
        else
        {
            // Enable the local keyboard repeat control
            m_check_lkbd_repeat->Enable(true);

            // If the feature is not enabled disable the sliders
            if (!m_check_lkbd_repeat->IsChecked())
            {
                init_lkbd_sliders(false);
            }
            else
            {
                // Setup the keyboard repeat delay slider
                m_slider_lkbd_delay->SetRange(0, LKBD_DELAY_NUM_TICS - 1);
                m_slider_lkbd_delay->SetTickFreq(LKBD_DELAY_INC_PER_TIC, 1);
                lkbd_delay_set_gui((UINT16)m_lkbd_delay);

                // Setup the keyboard repeat rate slider
                m_slider_lkbd_rate->SetRange(0, LKBD_RATE_NUM_TICS - 1);
                m_slider_lkbd_rate->SetTickFreq(LKBD_RATE_INC_PER_TIC, 1);
                lkbd_rate_set_gui((UINT16)m_lkbd_rate);
            }
        }

        // Need both OSD mouse speed set and get
        ret = wxGetApp().gui().pcoip_link().lcur_spd_avail();
        if (ret == TERA_SUCCESS)
        {
            ret = wxGetApp().gui().pcoip_link().lcur_spd_get(&lcur_spd);
        }
        if (ret != TERA_SUCCESS)
        {
            mTERA_LOG_DBG("local cursor speed not available");

            // Disable local cursor speed controls
            init_lcur_slider(false);
        }
        else
        {
            init_lcur_slider(true);

            // Set the slider range
            m_lcur_spd_slider->SetRange(10, 250);
            m_lcur_spd_slider->SetTickFreq(24, 1);

            // Set the slider to the current speed
            m_lcur_spd_slider->SetValue(lcur_spd);
        }

        // Update the controls
        lcur_lkbd_ctrl_update();
    } // end if
} // cPCOIP_AGENT_GUI_PROP_FEAT::init_dialog_lcur_lkbd()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lcur_lkbd_ctrl_update
//      Update the local cursor/keyboard controls
//
//  Description:
//      This function updates the local cursor/keyboard controls based on the
//      value of the enable checkbox. If the checkbox is checked, i.e., the
//      local cursor/keyboard is enabled, then the sub-options are enabled.
//      Otherwise, the sub-options are disabled (greyed out).
//
void cPCOIP_AGENT_GUI_PROP_FEAT::lcur_lkbd_ctrl_update(void)
{
    UINT32 lcur_vis;

    // Only update the controls if the enable checkbox is enabled
    if (m_lcur_lkbd_en_chk->IsChecked())
    {
        // Only enable the cursor visibility controls if available
        if (wxGetApp().gui().pcoip_link().lcur_shape_bmp_send_avail() == TERA_SUCCESS)
        {
            m_lcur_vis_txt->Enable();
            m_lcur_vis_choice->Enable();
        }

        // Only enable the speed controls if its available
        if (wxGetApp().gui().pcoip_link().lcur_spd_avail() == TERA_SUCCESS)
        {
            init_lcur_slider(true);
        }

        // Only enable the idle hide control if its available AND the
        // LOCAL cursor is visible
        lcur_vis = cur_visibility_get_reg();
        if ((wxGetApp().gui().pcoip_link().lcur_idle_hide_avail() == TERA_SUCCESS) &&
            (lcur_vis != PCOIP_AGENT_CUR_VISIBLE_HOST))
        {
            m_lcur_fade_chk->Enable();
        }
        else
        {
            m_lcur_fade_chk->Disable();
        }

        // Only enable the speed controls if its available
        if (wxGetApp().gui().pcoip_link().lcur_spd_avail() == TERA_SUCCESS)
        {
            init_lcur_slider(true);
        }

        // Only enable the keyboard repeat controls if its available
        if (wxGetApp().gui().pcoip_link().lkbd_client_auto_repeat_avail() == TERA_SUCCESS)
        {
            m_check_lkbd_repeat->Enable(true);
            init_lkbd_sliders(true);

            // Setup the keyboard repeat delay slider
            m_slider_lkbd_delay->SetRange(0, LKBD_DELAY_NUM_TICS - 1);
            m_slider_lkbd_delay->SetTickFreq(LKBD_DELAY_INC_PER_TIC, 1);
            lkbd_delay_set_gui((UINT16)m_lkbd_delay);

            // Setup the keyboard repeat rate slider
            m_slider_lkbd_rate->SetRange(0, LKBD_RATE_NUM_TICS - 1);
            m_slider_lkbd_rate->SetTickFreq(LKBD_RATE_INC_PER_TIC, 1);
            lkbd_rate_set_gui((UINT16)m_lkbd_rate);
        }
    }
    else
    {
        m_lcur_vis_txt->Disable();
        m_lcur_vis_choice->Disable();

        init_lcur_slider(false);

        m_lcur_fade_chk->Disable();

        m_check_lkbd_repeat->Enable(false);
        init_lkbd_sliders(false);
    }

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_FEAT::lcur_lkbd_ctrl_update()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_NET::help_btn_setup
//      Set up and show the tooltip help button
//
//  Description:
//      This function sets up and shows the local cursor/keyboard help button.
//      The button is formatted to look like plain text. When the user hovers
//      or clicks the text, a tooltip will appear. The tooltip is populated
//      with the reason why the local cursor/keyboard option is greyed out.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::lcur_hlp_btn_setup(int display_help)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    UINT32                                 stat;
//    sTERA_PCOIP_MSG_ENV_FIRM_MON_INFO      mon_info[TERA_PCOIP_MSG_MAX_NUM_MON];

    if (!display_help)
    {
        // Hide the help text
        m_help_link->Show(false);
    }
    else
    {
        // Show the help text and set it to something generic first
        m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_SUPPORT);

        // ...then determine the best help text if it exists
        if (!wxGetApp().gui().pcoip_link().is_ready())
        {
            mTERA_LOG_DBG("!is_ready");
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_LINK);
        }
        else if ((wxGetApp().gui().pcoip_link().sess_link_get(&sess_link) == TERA_SUCCESS) &&
                 !(sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
        {
            mTERA_LOG_DBG("no session");
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_PCOIP_SESS);
        }
        else if (!wxGetApp().gui().m_prop_sheet.m_mon_page->is_screen_cfg_ok())
        {
            mTERA_LOG_DBG("screen configuration data invalid");
            m_help_link->SetToolTip(IDS_STR_HELP_BAD_SCREEN_CFG);
        }
#if 0
        else if (wxGetApp().gui().pcoip_link().num_client_monitors() ==  0)
        {
            // use the same message as passing zero monitors to the client
            mTERA_LOG_ERR("zero monitors connected");
            m_help_link->SetToolTip(IDS_STR_HELP_LCUR_FF01);
        }
#endif
        else if (wxGetApp().gui().pcoip_link().lcur_lkbd_dev_feat_stat_get(&stat) == TERA_SUCCESS)
        {
            mTERA_LOG_DBG("local cursor status '0x%08X'", stat);

            if(stat == 0xFF00)
            {
                m_help_link->SetToolTip(IDS_STR_HELP_LCUR_FF00);
            }
            else if(stat == 0xFF01)
            {
                m_help_link->SetToolTip(IDS_STR_HELP_LCUR_FF01);
            }
            else if(stat == 0xFF02)
            {
                m_help_link->SetToolTip(IDS_STR_HELP_LCUR_FF02);
            }
            else if(stat == 0xFF03)
            {
                m_help_link->SetToolTip(IDS_STR_HELP_LCUR_FF03);
            }
            else if(((stat & 0x00FF) == 0x0001) ||
                    ((stat & 0x00FF) == 0x0002) ||
                    ((stat & 0x00FF) == 0x0003) ||
                    ((stat & 0x00FF) == 0x0004))
            {
                m_help_link->SetToolTip(IDS_STR_HELP_LCUR_0001);
            }
        }
        m_help_link->Show(true);
    }
} // cPCOIP_AGENT_GUI_PROP_FEAT::lcur_hlp_btn_setup()
// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_set_gui
//      Update the value of the GUIs keyboard repeat delay slider
//
//  Description:
//      This function updates the current value of the keyboard repeat delay
//      setting on the feature page. The value written to the slider is
//      adjusted by this function because values on the slider are reversed
//      so that the smallest number is on the right and the largest number is
//      on the left.
//
//      The slider value is also scaled by the constant
//      LKBD_DELAY_INC_PER_TIC. This is done to ensure the values selected by
//      the user are not every integer values in the range. This ensures the
//      user can only select the values (1250, 1000, 750, 500 and 250), which
//      is what the zero client OSD implemented.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_set_gui(UINT16 client_delay)
{
    INT32 gui_val;

    gui_val = (TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DELAY_MAX -
               client_delay) / LKBD_DELAY_INC_PER_TIC;
    mTERA_LOG_DBG("Client keyboard delay=%d => GUI keyboard delay=%d",
                  client_delay, gui_val);
    m_slider_lkbd_delay->SetValue(gui_val);
    m_lkbd_delay_slider_set = 1;

} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_set_gui()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_set_gui
//      Update the value of the GUIs keyboard repeat rate slider
//
//  Description:
//      This function updates the current value of the keyboard repeat rate
//      setting on the feature page. The value written to the slider is
//      adjusted by this function because values on the slider are reversed
//      so that the smallest number is on the right and the largest number is
//      on the left.
//
//      The slider value is also scaled by the constant
//      LKBD_RATE_INC_PER_TIC. This is done to ensure the values selected by
//      the user are not every integer values in the range. This ensures the
//      user can only select the values (750, 678, 606, ..., 30), which is
//      what the zero client OSD implemented.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_set_gui(UINT16 client_rate)
{
    INT32 gui_val;

    gui_val = (TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_RATE_MAX -
              client_rate) / LKBD_RATE_INC_PER_TIC;
    mTERA_LOG_DBG("Client keyboard rate=%d => GUI keyboard rate=%d",
                  client_rate, gui_val);
    m_slider_lkbd_rate->SetValue(gui_val);
    m_lkbd_rate_slider_set = 1;

} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_set_gui()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::client_kbd_settings_avail_set
//      Update the varaible that indicates client supports reading kbd settings
//
//  Description:
//      This function updates the current value of the variable that indicates
//      whether or not the client supports reading the keyboard repeat
//      settings. Only zero client devices running FW release 3.2.0 and higher
//      support this. If the client supports this feature the variable
//      m_client_kbd_setting_avail will be set equal to TERA_SUCCESS (0),
//      otherwise it will equal something else.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::client_kbd_settings_avail_set(INT32 status)
{

    mTERA_LOG_DBG("Setting m_client_kbd_setting_avail=%d", status);
    m_client_kbd_setting_avail = status;

} // cPCOIP_AGENT_GUI_PROP_FEAT::client_kbd_settings_avail_set()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_get_gui
//      Read and return the value of the GUIs keyboard repeat delay slider
//
//  Description:
//      This function retrieves the current value of the keyboard repeat delay
//      setting from the feature page. The value reported by the slider is
//      adjusted by this function because values on the slider are reversed
//      so that the smallest number is on the right and the largest number is
//      on the left.
//
//      The slider value is also scaled by the constant
//      LKBD_DELAY_INC_PER_TIC. This is done to ensure the values selected by
//      the user are not every integer values in the range. This ensures the
//      user can only select the values (1250, 1000, 750, 500 and 250), which
//      is what the zero client OSD implemented.
//
UINT16 cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_get_gui(void)
{
    INT32                                  gui_val;
    UINT16                                 lkbd_delay;
    UINT16                                 lkbd_rate;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lkbd_repeat_en_cfg;
    int                                    lkbd_repeat_delay_cfg;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    wxGetApp().gui().host_cfg().get_repeat_en_setting(sess_link.mac,
                                                      &lkbd_repeat_en_cfg);

    //
    // if the feature is enabled use the slider values, otherwise use the
    // setting that disables the feature on the client
    //
    if (lkbd_repeat_en_cfg)
    {
        // if the slider is initialized read the value from it
        if (m_lkbd_delay_slider_set)
        {
            //
            // the value of the slider is adjusted to deal with the problem
            // mentioned in the function header
            //
            gui_val = m_slider_lkbd_delay->GetValue();
            m_lkbd_delay = TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DELAY_MAX -
                           (gui_val * LKBD_DELAY_INC_PER_TIC);
        }
        else
        {
            //
            // if the client supports reading the kbd repeat delay get the
            // setting from the client, otherwise get the settings from the
            // registry
            //
            if (wxGetApp().gui().pcoip_link().lkbd_repeat_rates_get(&lkbd_delay, &lkbd_rate) == TERA_SUCCESS)
            {
                mTERA_LOG_DBG("using client keyboard repeat delay setting %d",
                              lkbd_delay);
            }
            else
            {
                wxGetApp().gui().host_cfg().get_repeat_delay_setting(sess_link.mac,
                                                                     &lkbd_repeat_delay_cfg);
                lkbd_delay = (UINT16)lkbd_repeat_delay_cfg;
                mTERA_LOG_DBG("using registry keyboard repeat delay setting %d",
                              lkbd_delay);
            }
            m_lkbd_delay = lkbd_delay;
        }
    }
    else
    {
        m_lkbd_delay = TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_DELAY;
    }
    mTERA_LOG_DBG("client keyboard delay=%d", m_lkbd_delay);

    return (UINT16)m_lkbd_delay;
} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_get_gui()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_get_gui
//      Read and return the value of the GUIs keyboard repeat rate slider
//
//  Description:
//      This function retrieves the current value of the keyboard repeat rate
//      setting from the feature page. The value reported by the slider is
//      adjusted by this function because values on the slider are reversed
//      so that the smallest number is on the right and the largest number is
//      on the left.
//
//      The slider value is also scaled by the constant
//      LKBD_RATE_INC_PER_TIC. This is done to ensure the values selected by
//      the user are not every integer values in the range. This ensures the
//      user can only select the values (750, 678, 606, ..., 30), which is
//      what the zero client OSD implemented.
//
UINT16 cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_get_gui(void)
{
    INT32                                  gui_val;
    UINT16                                 lkbd_delay;
    UINT16                                 lkbd_rate;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lkbd_repeat_en_cfg;
    int                                    lkbd_repeat_rate_cfg;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    wxGetApp().gui().host_cfg().get_repeat_en_setting(sess_link.mac,
                                                      &lkbd_repeat_en_cfg);

    //
    // if the feature is enabled use the slider values, otherwise use the
    // setting that disables the feature on the client
    //
    if (lkbd_repeat_en_cfg)
    {
        // if the slider is initialized read the value from it
        if (m_lkbd_rate_slider_set)
        {
            //
            // the value of the slider is adjusted to deal with the problem
            // mentioned in the function header
            //
            gui_val = m_slider_lkbd_rate->GetValue();
            m_lkbd_rate = TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_RATE_MAX -
                          (gui_val * LKBD_RATE_INC_PER_TIC);
        }
        else
        {
            //
            // if the client supports reading the kbd repeat rate get the
            // setting from the client, otherwise get the settings from the
            // registry
            //
            if (wxGetApp().gui().pcoip_link().lkbd_repeat_rates_get(&lkbd_delay, &lkbd_rate) == TERA_SUCCESS)
            {
                mTERA_LOG_DBG("using client keyboard repeat rate setting %d",
                              lkbd_rate);
            }
            else
            {
                wxGetApp().gui().host_cfg().get_repeat_rate_setting(sess_link.mac,
                                                                    &lkbd_repeat_rate_cfg);
                lkbd_rate = (UINT16)lkbd_repeat_rate_cfg;
                mTERA_LOG_DBG("using registry keyboard repeat rate setting %d",
                              lkbd_rate);
            }
            m_lkbd_rate = lkbd_rate;
        }
    }
    else
    {
        m_lkbd_rate = TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_RATE;
    }
    mTERA_LOG_DBG("client keyboard rate=%d", m_lkbd_rate);

    return (UINT16)m_lkbd_rate;
} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_get_gui()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lcur_spd_get
//      Read and return the value of the GUIs cursor speed slider
//
//  Description:
//      This function retrieves the current value of the cursor speed setting
//      from the feature page.
//
INT32 cPCOIP_AGENT_GUI_PROP_FEAT::lcur_spd_get(void)
{
	return m_lcur_spd_slider->GetValue();
} // cPCOIP_AGENT_GUI_PROP_FEAT::lcur_spd_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lcur_spd_enable_get
//      Read and return the enable state of the GUIs cursor speed slider
//
//  Description:
//      This function returns
//      from the feature page.
//
bool cPCOIP_AGENT_GUI_PROP_FEAT::lcur_spd_enable_get(void)
{
	return m_lcur_spd_slider->IsEnabled();
} // cPCOIP_AGENT_GUI_PROP_FEAT::lcur_spd_enable_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::cur_visibility_get_reg
//      Read and return the value of the cursor visibility from the registry
//
//  Description:
//      This function retrieves the current value of the cursor visibility
//      setting stored in the registry.
//
//      - if a registry entry exists the code validates the value. If the value
//        is not valid the function returns "host"
//      - if no registry entry exists the code uses the DEFAULT value
//
//  Note: The value in the registry is not updated.
//
UINT32 cPCOIP_AGENT_GUI_PROP_FEAT::cur_visibility_get_reg(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lcur_vis;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    // Get the local cursor visibility setting from registry
    wxGetApp().gui().host_cfg().get_lcur_vis_setting(sess_link.mac,
                                                     &lcur_vis);

    // Make sure the value read from the registry is valid
    if (lcur_vis >= PCOIP_AGENT_CUR_VISIBLE_NUM)
    {
        mTERA_LOG_ERR("Value %d assigned to vis cursor state because reg key %s is invalid, using 'Remote'",
                      PCOIP_AGENT_CUR_VISIBLE_HOST,
                      lcur_vis);
        lcur_vis = PCOIP_AGENT_CUR_VISIBLE_HOST;
    }

    return lcur_vis;

} // cPCOIP_AGENT_GUI_PROP_FEAT::cur_visibility_get_reg()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_get
//      Read and return the value of the current keyboard repeat delay setting
//
//  Description:
//      This function retrieves the current value of the keyboard repeat delay.
//
UINT16 cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_get(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lkbd_repeat_en_cfg;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    wxGetApp().gui().host_cfg().get_repeat_en_setting(sess_link.mac,
                                                      &lkbd_repeat_en_cfg);

    //
    // if the feature is enabled use m_lkbd_delay, otherwise use the setting
    // that disables the feature on the client
    //
    if (lkbd_repeat_en_cfg)
    {
        mTERA_LOG_DBG("client keyboard delay=%d", m_lkbd_delay);
        return (UINT16)m_lkbd_delay;
    }
    else
    {
        mTERA_LOG_DBG("repeat disabled - client keyboard delay=%d",
                      TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_DELAY);
        return TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_DELAY;
    }
} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_get
//      Read and return the value of the current keyboard repeat rate
//
//  Description:
//      This function retrieves the current value of the keyboard repeat rate.
//
UINT16 cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_get(void)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;
    int                                    lkbd_repeat_en_cfg;

    mTERA_MEM_ZERO(&sess_link, sizeof(sess_link));
    wxGetApp().gui().pcoip_link().sess_link_get(&sess_link);

    wxGetApp().gui().host_cfg().get_repeat_en_setting(sess_link.mac,
                                                      &lkbd_repeat_en_cfg);

    //
    // if the feature is enabled use m_lkbd_rate, otherwise use the setting
    // that disables the feature on the client
    //
    if (lkbd_repeat_en_cfg)
    {
        mTERA_LOG_DBG("client keyboard rate=%d", m_lkbd_rate);
        return (UINT16)m_lkbd_rate;
    }
    else
    {
        mTERA_LOG_DBG("repeat disabled - client keyboard rate=%d",
                      TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_RATE);
        return TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_RATE;
    }
} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_get()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_set
//      Updates the value of the current keyboard repeat delay setting
//
//  Description:
//      This function updates the current value of the keyboard repeat delay.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_set(UINT16 delay)
{
    m_lkbd_delay = delay;
} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_delay_set()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_set
//      Updates the value of the current keyboard repeat rate setting
//
//  Description:
//      This function updates the current value of the keyboard repeat rate.
//
void cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_set(UINT16 rate)
{
    m_lkbd_rate = rate;
} // cPCOIP_AGENT_GUI_PROP_FEAT::lkbd_rate_set()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::cPCOIP_AGENT_GUI_PROP_FEAT
///     Constructor
//
//  Description:
///     This is the PCoIP Host Software Features property page class
///     constructor. This function initializes the member data variables.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_FEAT::cPCOIP_AGENT_GUI_PROP_FEAT(
    wxWindow *parent
    ) :
    cPCOIP_AGENT_GUI_PROP_FEAT_BASE(parent)
{
    mTERA_LOG_DBG("start");

    // Initialize local variables
    m_lkbd_delay_slider_set = 0;
    m_lkbd_rate_slider_set = 0;
    m_client_kbd_setting_avail = TERA_ERR_TIMEOUT;
    m_lkbd_delay = TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_DELAY;
    m_lkbd_rate = TERA_PCOIP_HDC_APDU_KMP_KBD_AUTO_REPEAT_DISABLE_RATE;

    mTERA_LOG_DBG("lkbd delay=%d, rate=%d",
                  m_lkbd_delay,
                  m_lkbd_rate);
} // cPCOIP_AGENT_GUI_PROP_FEAT::cPCOIP_AGENT_GUI_PROP_FEAT()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_FEAT::~cPCOIP_AGENT_GUI_PROP_FEAT
///     Destructor
//
//  Description:
///     This is the PCoIP Host Software Features property page class
///     destructor. This function cleans up any loose ends before being
///     destroyed.
///
///     This destructor is called AFTER the property sheet's destructor.
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_FEAT::~cPCOIP_AGENT_GUI_PROP_FEAT(void)
{
    mTERA_LOG_DBG("start");

    m_lkbd_delay_slider_set = 0;
    m_lkbd_rate_slider_set = 0;

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_FEAT::~cPCOIP_AGENT_GUI_PROP_FEAT()


