////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Image Property Page Class Implementation
///
///     This file implements the PCoIP Host Software Image property page
///     class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include <wx/display.h>

#include "tera_log.h"
#include "tera_errors.h"
#include "tera_types.h"
#include "tera_mem.h"
#include "tera_pcoip_agent_internal.h"
#include "tera_pcoip_agent_gui_prop_img.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI_PROP_IMG, wxPanel)
    EVT_INIT_DIALOG(cPCOIP_AGENT_GUI_PROP_IMG::evt_init_dialog)
    EVT_BUTTON(wxID_APPLY_BTN, cPCOIP_AGENT_GUI_PROP_IMG::evt_apply_btn)
    EVT_COMMAND_SCROLL(wxID_ANY, cPCOIP_AGENT_GUI_PROP_IMG::evt_any_scroll)
END_EVENT_TABLE()

//
// the following constants are used to scale the image quality preference
// slider settings to be compatible with the settings used by the client device
//
#define IMG_QUAL_PREF_NUM_TICS     21
#define IMG_QUAL_PREF_RANGE        (TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MAX - \
                                    TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MIN)
#define IMG_QUAL_PREF_INC_PER_TIC  (IMG_QUAL_PREF_RANGE / (IMG_QUAL_PREF_NUM_TICS - 1))
#define IMG_QUAL_PREF_INVALID      (TERA_PCOIP_HCC_APDU_IMG_QUAL_PREF_MAX + 1)

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::evt_any_scroll
//      Process the scroll change event
//
//  Description:
//      This function processes scroll change events. It reads the value of
//      the slider and writes the setting to the value to the box that stores
//      the digital value of the slider. It then enables the Apply button,
//      which allows teh user to activate the new setting. Activating the
//      setting involves writing the setting to the host.
//
void cPCOIP_AGENT_GUI_PROP_IMG::evt_any_scroll(
    wxScrollEvent& evt
    )
{
    UINT32 gui_img_qual;

    gui_img_qual = img_qual_pref_get_gui();
    img_qual_pref_set_gui(gui_img_qual);

    // Enable the apply button
    m_apply_btn->Enable(1);
} // cPCOIP_AGENT_GUI_PROP_FEAT::evt_any_scroll()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::evt_init_dialog
//      Process the evt_init_dialog event
//
//  Description:
//      This function processes the init dialog event.
//
void cPCOIP_AGENT_GUI_PROP_IMG::evt_init_dialog(
    wxInitDialogEvent& evt
    )
{
    sTERA_PCOIP_LINK_IMG_USER_CFG   user_cfg;

    mTERA_LOG_DBG("IMG EVT_INIT_DIALOG");

    if (wxGetApp().gui().pcoip_link().img_user_cfg_req_get(&user_cfg) != TERA_SUCCESS)
    {
        mTERA_LOG_DBG("image quality pref not available");

        init_img_slider(FALSE);
        help_btn_setup(1);
    }
    else
    {
        // Setup the image quality preference slider
        init_img_slider(TRUE);
        m_img_qual_pref_slider->SetRange(0, IMG_QUAL_PREF_NUM_TICS - 1);
        m_img_qual_pref_slider->SetTickFreq(1, 1);
        m_img_qual_pref = user_cfg.image_qual_pref;
        img_qual_pref_set_gui(m_img_qual_pref);
        help_btn_setup(0);
    }

    evt.Skip();
} // cPCOIP_AGENT_GUI_PROP_IMG::evt_init_dialog()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::evt_apply_btn
//      Process the apply button click event
//
//  Description:
//      This function calls the function that saves and applies the current
//      settings.
//
void cPCOIP_AGENT_GUI_PROP_IMG::evt_apply_btn(
    wxCommandEvent& evt
    )
{
    UINT32 gui_img_qual;

    gui_img_qual = img_qual_pref_get_gui();

    // update the GUI and send new settings to the FW if the image quality
    // setting is valid
    if (gui_img_qual != IMG_QUAL_PREF_INVALID)
    {
        img_qual_pref_set_gui(gui_img_qual);

        // Send user config cache set apdu
        wxGetApp().gui().pcoip_link().img_user_cfg_set(gui_img_qual);
        wxGetApp().gui().pcoip_link().img_user_cfg_cache_set(gui_img_qual);

        mTERA_LOG_DBG("end: gui img qual = %d", gui_img_qual);
    }

    // dsiable the Apply button
    m_apply_btn->Enable(0);
} // cPCOIP_AGENT_GUI_PROP_IMG::evt_apply_btn()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::init_img_slider
//      Initialize the image quality preference slider
//
//  Description:
//      This function initializes the image quality preference slider and
//      value box.
//
void cPCOIP_AGENT_GUI_PROP_IMG::init_img_slider(bool enable)
{
    m_img_qual_pref_slider->Enable(enable);
    m_img_qual_pref_slider->SetRange(0, IMG_QUAL_PREF_NUM_TICS-1);
    m_img_qual_pref_slider->SetTickFreq(1, 1);
    m_img_qual_sharp_txt->Enable(enable);
    m_img_qual_smooth_txt->Enable(enable);
    m_img_qual_pref_txt->Enable(enable);
    m_img_qual_pref_val->Enable(enable);
    m_apply_btn->Enable(0);
} // cPCOIP_AGENT_GUI_PROP_IMG::init_img_slider()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::img_qual_pref_set_gui
//      Update the value of the GUIs image quality preference slider
//
//  Description:
//      This function updates the current value of the image quality preference
//      setting on the image page. The value written to the slider is scaled
//      by the constant IMG_QUAL_PREF_INC_PER_TIC. This is done to ensure the
//      values selected by the user are not all integer values in the range.
//      This ensures the user can only select the values (0, 5, 10... 100).
//
void cPCOIP_AGENT_GUI_PROP_IMG::img_qual_pref_set_gui(UINT32 client_img_qual)
{
    INT32    gui_val;
    wxString val;

    gui_val = client_img_qual / IMG_QUAL_PREF_INC_PER_TIC;
    mTERA_LOG_DBG("Client img qual pref=%d => GUI img qual pref=%d",
                  client_img_qual, gui_val);
	m_img_qual_pref_slider->SetValue(gui_val);
    m_img_qual_pref_slider_set = 1;
    val.Printf(_T("%02d"), client_img_qual);
    m_img_qual_pref_val->SetValue(val);
} // cPCOIP_AGENT_GUI_PROP_IMG::img_qual_pref_set_gui()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::img_qual_pref_get_gui
//      Read and return the value of the GUIs keyboard repeat delay slider
//
//  Description:
//      This function retrieves the current value of the image quality
//      preference setting from the image page. The value reported by the
//      slider is scaled by the constant IMG_QUAL_PREF_INC_PER_TIC. This is
//      done to ensure the values selected by the user are not every integer
//      values in the range. This ensures the user can only select the values
//      (0, 10, 20, ... 100).
//
UINT32 cPCOIP_AGENT_GUI_PROP_IMG::img_qual_pref_get_gui(void)
{
    INT32                           gui_val;
    sTERA_PCOIP_LINK_IMG_USER_CFG   user_cfg;

    // if the slider is initialized read the value from it
    if (m_img_qual_pref_slider_set)
    {
        //
        // the value of the slider is adjusted to deal with the problem
        // mentioned in the function header
        //
        gui_val = m_img_qual_pref_slider->GetValue();
        m_img_qual_pref = gui_val * IMG_QUAL_PREF_INC_PER_TIC;
    }
    else
    {
        //
        // if the client supports reading the image quality preference get the
        // setting from the client, otherwise disable the slider(?)
        //
        if (wxGetApp().gui().pcoip_link().img_user_cfg_req_get(&user_cfg) != TERA_SUCCESS)
        {
            m_img_qual_pref = IMG_QUAL_PREF_INVALID;
            mTERA_LOG_DBG("FW doesn't support image quality settings");
        }
        else
        {
            m_img_qual_pref = user_cfg.image_qual_pref;
        }
    }

    mTERA_LOG_DBG("client image qual pref=%d", m_img_qual_pref);

    return m_img_qual_pref;
} // cPCOIP_AGENT_GUI_PROP_IMG::img_qual_pref_get_gui()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::help_btn_setup
//      Set up and show the tooltip help button
//
//  Description:
//      This function sets up and shows the tooltip help button. The button
//      is formatted to look like plain text. When the user hovers or clicks
//      the text, a tooltip will appear. The tooltip provides possible
//      causes and solutions to why certain features are unavailable.
//
void cPCOIP_AGENT_GUI_PROP_IMG::help_btn_setup(bool display_help)
{
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK sess_link;

    if (!display_help)
    {
        // Hide the help text
        m_help_link->Show(false);
    }
    else
    {
        // Show the help text and set it to something generic first
        m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_SUPPORT);

        // ...then determine the best help text if it exists
        if (!wxGetApp().gui().pcoip_link().is_ready())
        {
            mTERA_LOG_DBG("!is_ready");
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_LINK);
        }
        else if ((wxGetApp().gui().pcoip_link().sess_link_get(&sess_link) == TERA_SUCCESS) &&
                 !(sess_link.flags & TERA_PCOIP_HCC_APDU_HMI_VAR_SESS_LINK_FLAGS_CONNECTED))
        {
            mTERA_LOG_DBG("no session");
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_PCOIP_SESS);
        }
        m_help_link->Show(true);
    }
} // cPCOIP_AGENT_GUI_PROP_IMG::help_btn_setup()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::cPCOIP_AGENT_GUI_PROP_IMG
///     Constructor
//
//  Description:
///     This is the PCoIP Host Software image property page class constructor.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_IMG::cPCOIP_AGENT_GUI_PROP_IMG(
    wxWindow *parent
    ) :
    cPCOIP_AGENT_GUI_PROP_IMG_BASE(parent)
{
    // Initialize local variables
    m_img_qual_pref = IMG_QUAL_PREF_INVALID;
    m_img_qual_pref_slider_set = 0;

} // cPCOIP_AGENT_GUI_PROP_IMG::cPCOIP_AGENT_GUI_PROP_IMG()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_IMG::~cPCOIP_AGENT_GUI_PROP_IMG
///     Destructor
//
//  Description:
///     This is the PCoIP Host Software image property page class destructor.
///     This function cleans up any loose ends before being destroyed.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_IMG::~cPCOIP_AGENT_GUI_PROP_IMG(void)
{
    m_img_qual_pref_slider_set = 0;
} // cPCOIP_AGENT_GUI_PROP_IMG::~cPCOIP_AGENT_GUI_PROP_IMG()
