////////////////////////////////////////////////////////////////////////////
//   Copyright 2013  Teradici Corporation
//
//   This program is free software; you can redistribute it and/or modify it
//   under the terms and conditions of the GNU General Public License,
//   version 2, as published by the Free Software Foundation.
//
////////////////////////////////////////////////////////////////////////////
/// \file
/// \brief
///     PCoIP Host Software Statistics Property Page Class Implementation
///
///     This file implements the PCoIP Host Software Statistics property page
///     class.
///
////////////////////////////////////////////////////////////////////////////

// ---- Include Files -------------------------------------------------------

#include <wx/wxprec.h>

#if !defined(WX_PRECOMP)
    #include <wx/wx.h>
#endif

#include "tera_errors.h"
#include "tera_log.h"

#include "tera_pcoip_agent_gui_prop_stat.h"
#include "tera_pcoip_agent_internal.h"

// ---- Global Variables ----------------------------------------------------

// ---- Private Constants and Types -----------------------------------------

// This constant determines how often the statistics are updated (in ms)
#define STATISTICS_UPDATE_INTERVAL_MS   1000

#define BW_MIN_INVALID_VAL  (1000)
#define BW_MAX_INVALID_VAL  (0)

// This constant determines how many times an error can occur reading the
// session statistics before the agent stops reading the statistics
#define MAX_STATS_GET_FAIL  100

BEGIN_EVENT_TABLE(cPCOIP_AGENT_GUI_PROP_STAT, wxPanel)
    EVT_TIMER(wxID_ANY, cPCOIP_AGENT_GUI_PROP_STAT::evt_timer)
    EVT_BUTTON(wxID_RESET_BTN, cPCOIP_AGENT_GUI_PROP_STAT::evt_reset_btn)
    EVT_INIT_DIALOG(cPCOIP_AGENT_GUI_PROP_STAT::evt_init_dialog)
    EVT_CLOSE(cPCOIP_AGENT_GUI_PROP_STAT::evt_close)
END_EVENT_TABLE()

// ---- Local Variables -----------------------------------------------------

// ---- Local Function Prototypes -------------------------------------------

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::evt_timer
//      Process the timer firing off event
//
//  Description:
//      This function process the timer firing off event. This function updates
//      property grid with the latest session statistics.
//
void cPCOIP_AGENT_GUI_PROP_STAT::evt_timer(
    wxTimerEvent& evt
    )
{
    if (TERA_SUCCESS != prop_grid_update())
    {
        mTERA_LOG_ERR("failed to update session statistics");

        // keep track of how many times this failure occurred
        m_get_stat_fail_cnt++;

        if (m_get_stat_fail_cnt > MAX_STATS_GET_FAIL)
        {
            // Session statistics are unavailable, so stop the timer
            m_update_timer.Stop();
            mTERA_LOG_ERR("Failed to get session statistics %d times, giving up",
                          m_get_stat_fail_cnt);
        }
    }
} // cPCOIP_AGENT_GUI_PROP_STAT::evt_timer()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::evt_reset_btn
//      Process the Reset Statistics button click event
//
//  Description:
//
//
void cPCOIP_AGENT_GUI_PROP_STAT::evt_reset_btn(
    wxCommandEvent& evt
    )
{
    // Reset the session statistics
    wxGetApp().gui().pcoip_link().sess_stat_reset();

    // Reset all the property grid values
    prop_grid_reset();
} // cPCOIP_AGENT_GUI_PROP_STAT::evt_reset_btn()

void cPCOIP_AGENT_GUI_PROP_STAT::evt_close(
    wxCloseEvent& evt
    )
{
    mTERA_LOG_DBG("start - cPCOIP_AGENT_GUI_PROP_STAT");

    evt.Skip();

    mTERA_LOG_DBG("end - cPCOIP_AGENT_GUI_PROP_STAT");
} // cPCOIP_AGENT_GUI_PROP_STAT::evt_close()

void cPCOIP_AGENT_GUI_PROP_STAT::evt_init_dialog(
    wxInitDialogEvent& evt
    )
{
    mTERA_LOG_DBG("STAT EVT_INIT_DIALOG");

    evt.Skip();

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_STAT::evt_init_dialog()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_populate
//      Populate the property grid
//
//  Description:
//      This function populates the property grid with all the entries and
//      assigns all fields their default values.
//
void cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_populate(void)
{
    m_prop_grid->Append(new wxPropertyCategory(_("Traffic")));

    m_prop_grid->Append(new wxStringProperty(_("PCoIP Packets Sent"),
                                             _T("tx_pkt_cnt"), _T("0")));
    m_prop_grid->Append(new wxStringProperty(_("PCoIP Packets Received"),
                                             _T("rx_pkt_cnt"), _T("0")));
    m_prop_grid->Append(new wxStringProperty(_("PCoIP Packets Lost"),
                                             _T("rx_pkt_loss_cnt"), _T("0")));
    m_prop_grid->Append(new wxStringProperty(_("Bytes Sent"),
                                             _T("tx_byte_cnt"), _T("0")));
    m_prop_grid->Append(new wxStringProperty(_("Bytes Received"),
                                             _T("rx_byte_cnt"), _T("0")));

    m_prop_grid->Append(new wxPropertyCategory(_("Round Trip Latency")));

    m_prop_grid->Append(new wxStringProperty(_("Min / Current / Max Latency"),
                                             _T("latency"),
                                             _T("0 / 0 / 0 ms")));

    m_prop_grid->Append(new wxPropertyCategory(_("Bandwidth")));

    m_prop_grid->Append(new wxStringProperty(_("Active Limit"),
                                             _T("bw_active_limit"),
                                             _T("0.00 kbps")));
    m_prop_grid->Append(new wxStringProperty(_("Min / Current / Max Utilization"),
                                             _T("bw"),
                                             _T("0.00 / 0.00 / 0.00 kbps")));

    m_prop_grid->Append(new wxPropertyCategory(_("Imaging")));

    m_prop_grid->Append(new wxStringProperty(_("Display 1 Frame Rate"),
                                             _T("frame_rate_0"), _T("0 fps")));
    m_prop_grid->Append(new wxStringProperty(_("Display 2 Frame Rate"),
                                             _T("frame_rate_1"), _T("0 fps")));

    if (4 == m_max_disp)
    {
        m_prop_grid->Append(new wxStringProperty(_("Display 3 Frame Rate"),
                                                 _T("frame_rate_2"), _T("0 fps")));
        m_prop_grid->Append(new wxStringProperty(_("Display 4 Frame Rate"),
                                                 _T("frame_rate_3"), _T("0 fps")));
    }

    // Fit first column to width of longest label
    m_prop_grid->FitColumns();
} // cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_populate()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_reset
//      Reset all the property grid values
//
//  Description:
//      This function resets all the property grid values to zero.
//
void cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_reset(void)
{
    m_prop_grid->SetPropertyValue(_T("tx_pkt_cnt"), _T("0"));
    m_prop_grid->SetPropertyValue(_T("rx_pkt_cnt"), _T("0"));
    m_prop_grid->SetPropertyValue(_T("rx_pkt_loss_cnt"), _T("0"));
    m_prop_grid->SetPropertyValue(_T("tx_byte_cnt"), _T("0"));
    m_prop_grid->SetPropertyValue(_T("rx_byte_cnt"), _T("0"));

    m_prop_grid->SetPropertyValue(_T("latency"), _T("0 / 0 / 0 ms"));

    m_prop_grid->SetPropertyValue(_T("bw_active_limit"), _T("0.00 kbps"));
    m_prop_grid->SetPropertyValue(_T("bw"), _T("0.00 / 0.00 / 0.00 kbps"));

    m_prop_grid->SetPropertyValue(_T("frame_rate_0"), _T("0 fps"));
    m_prop_grid->SetPropertyValue(_T("frame_rate_1"), _T("0 fps"));

    if (4 == m_max_disp)
    {
        m_prop_grid->SetPropertyValue(_T("frame_rate_2"), _T("0 fps"));
        m_prop_grid->SetPropertyValue(_T("frame_rate_3"), _T("0 fps"));
    }
    // Reset the min and max bandwidth values
    m_bw_min = BW_MIN_INVALID_VAL;
    m_bw_max = BW_MAX_INVALID_VAL;

    // Reset the reference Tx byte count
    m_last_tx_byte_cnt = 0;
} // cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_reset()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_update
//      Update the contents of the property grid control
//
//  Description:
//      This function updates the contents of the property grid control with
//      the latest session statistics if the GUI is active.

//      If the session statistics are not available, then this function will
//      disable all the fields.
//
INT32 cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_update(void)
{
    INT32                                  ret;
    sTERA_PCOIP_HCC_APDU_HMI_VAR_SESS_STAT sess_stat;
    wxString                               val;
    int                                    display_help = 0;

    // don't update the stats when the GUI is closed
    // TBD - It would be nice if the code only retrieved the statistics
    //       when the stats page is active, currently it gets the stats
    //       when the GUI is open.
    if (!wxGetApp().gui().prop_sheet_is_shown())
    {
        return TERA_SUCCESS;
    }

    // Get the session statistics
    ret = wxGetApp().gui().pcoip_link().sess_stat_get(&sess_stat);
    if (ret == TERA_SUCCESS)
    {
        val.Printf(_T("%u"), sess_stat.tx_pkt_cnt);
        m_prop_grid->SetPropertyValue(_T("tx_pkt_cnt"), val);
        val.Printf(_T("%u"), sess_stat.rx_pkt_cnt);
        m_prop_grid->SetPropertyValue(_T("rx_pkt_cnt"), val);
        val.Printf(_T("%u"), sess_stat.rx_pkt_loss_cnt);
        m_prop_grid->SetPropertyValue(_T("rx_pkt_loss_cnt"), val);
        val.Printf(_T("%u"), sess_stat.tx_byte_cnt);
        m_prop_grid->SetPropertyValue(_T("tx_byte_cnt"), val);
        val.Printf(_T("%u"), sess_stat.rx_byte_cnt);
        m_prop_grid->SetPropertyValue(_T("rx_byte_cnt"), val);

        val.Printf(_T("%u / %u / %u ms"), sess_stat.latency_min,
                    sess_stat.latency_cur, sess_stat.latency_max);
        m_prop_grid->SetPropertyValue(_T("latency"), val);

        //
        // Detect when a session has gone from disconnect->connect by comparing
        // the current Tx byte count to the previous value we have. The values
        // are reset automatically in the driver when a session goes from
        // disconnect->connect.
        //
        if (sess_stat.tx_byte_cnt < m_last_tx_byte_cnt)
        {
            // Reset the min and max bandwidth values
            m_bw_min = BW_MIN_INVALID_VAL;
            m_bw_max = BW_MAX_INVALID_VAL;
        }

        // Record the Tx byte count for reference
        m_last_tx_byte_cnt = sess_stat.tx_byte_cnt;

        if (sess_stat.tx_bw_avg < m_bw_min)
        {
            m_bw_min = sess_stat.tx_bw_avg;
        }
        else if (sess_stat.tx_bw_avg > m_bw_max)
        {
            m_bw_max = sess_stat.tx_bw_avg;
        }

        val.Printf(_T("%u kbps"), sess_stat.tx_bw_limit);
        m_prop_grid->SetPropertyValue(_T("bw_active_limit"), val);
        val.Printf(_T("%u / %u / %u kbps"), sess_stat.tx_bw_min,
                   sess_stat.tx_bw_avg, sess_stat.tx_bw_max);
        m_prop_grid->SetPropertyValue(_T("bw"), val);

        val.Printf(_T("%u fps"), sess_stat.frame_rate[0]);
        m_prop_grid->SetPropertyValue(_T("frame_rate_0"), val);
        val.Printf(_T("%u fps"), sess_stat.frame_rate[1]);
        m_prop_grid->SetPropertyValue(_T("frame_rate_1"), val);
        if (4 == m_max_disp)
        {
            val.Printf(_T("%u fps"), sess_stat.frame_rate[2]);
            m_prop_grid->SetPropertyValue(_T("frame_rate_2"), val);
            val.Printf(_T("%u fps"), sess_stat.frame_rate[3]);
            m_prop_grid->SetPropertyValue(_T("frame_rate_3"), val);
        }
    }
    else
    {
        // Disable all the values
        m_prop_grid->SetPropertyValue(_T("tx_pkt_cnt"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("rx_pkt_cnt"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("rx_pkt_loss_cnt"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("tx_byte_cnt"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("rx_byte_cnt"), _("Unavailable"));

        m_prop_grid->SetPropertyValue(_T("latency"), _("Unavailable"));

        m_prop_grid->SetPropertyValue(_T("bw_active_limit"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("bw"), _("Unavailable"));

        m_prop_grid->SetPropertyValue(_T("frame_rate_0"), _("Unavailable"));
        m_prop_grid->SetPropertyValue(_T("frame_rate_1"), _("Unavailable"));
        if (4 == m_max_disp)
        {
            m_prop_grid->SetPropertyValue(_T("frame_rate_2"), _("Unavailable"));
            m_prop_grid->SetPropertyValue(_T("frame_rate_3"), _("Unavailable"));
        }

        // Disable the reset button
        m_reset_btn->Enable(false);

        // Display the help tooltips
        display_help = 1;
    }

    // Setup and show the help tooltips
    help_btn_setup(display_help);

    return ret;
} // cPCOIP_AGENT_GUI_PROP_STAT::prop_grid_update()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::help_btn_setup
//      Set up and show the tooltip help button
//
//  Description:
//      This function sets up and shows the tooltip help button. The button
//      is formatted to look like plain text. When the user hovers or clicks
//      the text, a tooltip will appear. The tooltip provides possible
//      causes and solutions to why certain features are unavailable.
//
void cPCOIP_AGENT_GUI_PROP_STAT::help_btn_setup(int display_help)
{
    if (!display_help)
    {
        // Hide the help text
        m_help_link->Show(false);
    }
    else
    {
        // Show the help text and set it to something generic first
        m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_SUPPORT);

        // ...then determine the best help text if it exists
        if (!wxGetApp().gui().pcoip_link().is_ready())
        {
            m_help_link->SetToolTip(IDS_STR_HELP_GEN_NO_LINK);
        }
        else if (m_get_stat_fail_cnt > MAX_STATS_GET_FAIL)
        {
            m_help_link->SetToolTip(IDS_STR_HELP_STAT_GET_ERR);
        }
        m_help_link->Show(true);
    }
} // cPCOIP_AGENT_GUI_PROP_STAT::help_btn_setup()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::cPCOIP_AGENT_GUI_PROP_STAT
///     Constructor
//
//  Description:
///     This is the Statistics property page class constructor.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_STAT::cPCOIP_AGENT_GUI_PROP_STAT(
    wxWindow *parent
    ) :
    cPCOIP_AGENT_GUI_PROP_STAT_BASE(parent),
    m_update_timer(this)
{
    mTERA_LOG_DBG("start");

    // Initialize variables
    m_bw_min = BW_MIN_INVALID_VAL;
    m_bw_max = BW_MAX_INVALID_VAL;
    m_last_tx_byte_cnt = 0;
    m_get_stat_fail_cnt = 0;

    // This variable is re-initialized later when the device_type is known
    m_max_disp = 2;

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_STAT::cPCOIP_AGENT_GUI_PROP_STAT()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::~cPCOIP_AGENT_GUI_PROP_STAT
///     Destructor
//
//  Description:
///     This is the Session property page class destructor. This function
///     cleans up any loose ends before being destroyed.
///
/// \return
///     None
///
cPCOIP_AGENT_GUI_PROP_STAT::~cPCOIP_AGENT_GUI_PROP_STAT(void)
{
    mTERA_LOG_DBG("start");

    // Stop the update timer
    if (m_update_timer.IsRunning())
    {
        m_update_timer.Stop();
    }

    mTERA_LOG_DBG("end");
} // cPCOIP_AGENT_GUI_PROP_STAT::~cPCOIP_AGENT_GUI_PROP_STAT()

// --------------------------------------------------------------------------
//  cPCOIP_AGENT_GUI_PROP_STAT::max_monitors_set
///     Set the maximum number of monitors
//
//  Description:
///     This function is called by an external module that knows the maximum
///     number of monitors supported by the PCoIP host card. It updates the
///     variable that keeps track of the maximum number of supported monitors
///     and it also populates the property grid, since the max monitors is
///     now known.
///
/// \return
///     None
///
void cPCOIP_AGENT_GUI_PROP_STAT::max_monitors_set(UINT32 max_monitors)
{
    m_max_disp = max_monitors;

    // Populate the property grid
    prop_grid_populate();

    // Start the update timer
    m_update_timer.Start(STATISTICS_UPDATE_INTERVAL_MS, wxTIMER_CONTINUOUS);
} // cPCOIP_AGENT_GUI_PROP_STAT::max_monitors_set()

